import logging
import tkinter as tk

import docutils.core
import docutils.nodes

from thonny import get_workbench, ui_utils
from thonny.codeview import get_syntax_options_for_tag
from thonny.tktextext import TweakableText


class RstText(TweakableText):
    
    def __init__(self, master=None, cnf={}, read_only=False, **kw):
        
        super().__init__(master=master, cnf=cnf, read_only=read_only, 
                         **{"font" : "TkDefaultFont",
                            #"cursor" : "",
                            **kw})
        self.configure_tags()
        self._visitor = None
    
    def configure_tags(self):
        main_font = tk.font.nametofont("TkDefaultFont")
        
        bold_font = main_font.copy()
        bold_font.configure(weight="bold", size=main_font.cget("size"))
        
        italic_font = main_font.copy()
        italic_font.configure(slant="italic", size=main_font.cget("size"))
        
        h1_font = main_font.copy()
        h1_font.configure(size=main_font.cget("size") * 2, weight="bold")
        
        h2_font = main_font.copy()
        h2_font.configure(size=round(main_font.cget("size") * 1.5), weight="bold")
        
        h3_font = main_font.copy()
        h3_font.configure(size=main_font.cget("size"), weight="bold")
        
        self.tag_configure("h1", font=h1_font, spacing3=5)
        self.tag_configure("h2", font=h2_font, spacing3=5)
        self.tag_configure("h3", font=h3_font, spacing3=5)
        self.tag_configure("p", spacing1=0, spacing3=10, spacing2=0)
        self.tag_configure("line_block", spacing1=0, spacing3=10, spacing2=0)
        self.tag_configure("em", font=italic_font)
        self.tag_configure("strong", font=bold_font)
        self.tag_configure("a", **get_syntax_options_for_tag("hyperlink"))
        self.tag_bind("a", "<Enter>", self._hyperlink_enter)
        self.tag_bind("a", "<Leave>", self._hyperlink_leave)
        
        self.tag_configure("topic_title", font=bold_font, lmargin2=16)
        self.tag_configure("topic_body", lmargin1=16, lmargin2=16)
        self.tag_configure("code", font="TkFixedFont", 
                           #background="#eeeeee"
                           )
        #if ui_utils.get_tk_version_info() >= (8,6,6):
        #    self.tag_configure("code", lmargincolor=self["background"])
        
        for i in range(1,6):
            self.tag_configure("list%d" % i, lmargin1=i*10, lmargin2=i*10+10)
        
        toti_code_font = bold_font.copy()
        toti_code_font.configure(family=tk.font.nametofont("TkFixedFont").cget("family"),
                                 size=bold_font.cget("size"))
        self.tag_configure("topic_title_code", font=toti_code_font)
        self.tag_raise("topic_title_code", "code")
        self.tag_raise("topic_title_code", "topic_title")
        
        # TODO: topic_title + em
        self.tag_raise("em", "topic_title")
        
        if ui_utils.get_tk_version_info() >= (8,6,6):
            self.tag_configure("sel", lmargincolor=self["background"])
        self.tag_raise("sel")
    
    def clear(self):
        self.direct_delete("1.0", "end")
    
    def load_rst(self, rst_source, global_tags=()):
        self.clear()
        self.append_rst(rst_source, global_tags)
    
    def append_rst(self, rst_source, global_tags=()):
        doc = docutils.core.publish_doctree(rst_source)
        doc.walkabout(self.create_visitor(doc, global_tags))
    
        # For debugging:
        #self.direct_insert("end", doc.pformat())
        #self.direct_insert("end", rst_source)
    
    def create_visitor(self, doc, global_tags=()):
        # Pass unique tag count from previous visitor
        # to keep uniqueness
        
        if self._visitor is None:
            unique_tag_count = 0
        else:
            unique_tag_count = self._visitor.unique_tag_count
            
        self._visitor = TkTextRenderingVisitor(doc, self, global_tags, unique_tag_count)
        
        return self._visitor
        
    
    def _hyperlink_enter(self, event):
        self.config(cursor="hand2")
        
    def _hyperlink_leave(self, event):
        self.config(cursor="xterm")

class TkTextRenderingVisitor(docutils.nodes.GenericNodeVisitor):
    
    def __init__(self, document, text, global_tags=(), unique_tag_count=0):
        super().__init__(document)
        
        self._context_tags = list(global_tags)
        self.text = text
        self.section_level = 0
        self.in_topic = False
        self.in_paragraph = False
        self.in_title = False
        
        self.active_lists = []
        
        self.unique_tag_count = unique_tag_count
    
    def visit_document(self, node):
        pass
    
    def visit_Text(self, node):
        self._append_text(self._node_to_text(node))
    
    def visit_section(self, node):
        self.section_level += 1
    def depart_section(self, node):
        self.section_level -= 1
    
    def _get_title_tag(self):
        if self.in_topic:
            return "topic_title"
        else:
            return "h%d" % (self.section_level+1)
    
    def visit_title(self, node):
        self.in_title = True
        self._add_tag(self._get_title_tag())
    def depart_title(self, node):
        self.in_title = False
        self._append_text("\n")
        self._pop_tag(self._get_title_tag())
    
    def visit_paragraph(self, node):
        self.in_paragraph = True
        if not self.active_lists:
            self._add_tag("p")
    def depart_paragraph(self, node):
        self.in_paragraph = False
        self._append_text("\n")
        if not self.active_lists:
            self._pop_tag("p")
    
    def visit_line_block(self, node):
        self._add_tag("line_block")
    def depart_line_block(self, node):
        self._pop_tag("line_block")

    def visit_line(self, node):
        pass
    def depart_line(self, node):
        self._append_text("\n")
    
    def visit_topic(self, node):
        self.in_topic = True
        
        if "toggle" in node.attributes["classes"]:
            return self._visit_toggle_topic(node)
        else:
            return self.default_visit(node)
    
    def _visit_toggle_topic(self, node):
        tag = self._create_unique_tag()
        title_id_tag = tag + "_title"
        body_id_tag = tag + "_body"
        
        def get_toggler_image_name(kind):
            if get_workbench().uses_dark_ui_theme():
                return kind + "_light"
            else:
                return kind
        
        if "open" in node.attributes["classes"]:
            initial_image = get_toggler_image_name("boxminus")
            initial_elide = False
        else:
            initial_image = get_toggler_image_name("boxplus")
            initial_elide = True
        
        label = tk.Label(self.text,
                         image=get_workbench().get_image(initial_image),
                         borderwidth=0,
                         background=self.text["background"],
                         cursor="arrow")
        
        def toggle_body(event=None):
            elide = self.text.tag_cget(body_id_tag, "elide")
            if elide == '1':
                elide = True
            elif elide == '0':
                elide = False
            else:
                elide = bool(elide)
            
            elide = not elide
            
            self.text.tag_configure(body_id_tag, elide=elide)
            if self.text.has_selection():
                self.text.tag_remove("sel", "1.0", "end")
            
            if elide:
                label.configure(image=get_workbench()
                                .get_image(get_toggler_image_name("boxplus")))
            else:
                label.configure(image=get_workbench()
                                .get_image(get_toggler_image_name("boxminus")))
        
        assert isinstance(node.children[0], docutils.nodes.title)
        
        #self.text.tag_bind(title_id_tag, "<1>", toggle_body, True)
        self._add_tag(title_id_tag)
        self._append_window(label)
        label.bind("<1>", toggle_body, True)
        node.children[0].walkabout(self)
        self._pop_tag(title_id_tag)
        
        self.text.tag_configure(body_id_tag, elide=initial_elide)
        self._add_tag(body_id_tag)
        self._add_tag("topic_body")
        for child in list(node.children)[1:]:
            child.walkabout(self)
        self._pop_tag("topic_body")
        self._pop_tag(body_id_tag)
        
        if "tight" not in node.attributes["classes"]:
            self._append_text("\n")
        
        raise docutils.nodes.SkipNode()
            
    def depart_topic(self, node):
        # only for non-toggle topics
        self.in_topic = False
        self._append_text("\n")
    
    def visit_image(self, node):
        self._append_image(node.attributes["uri"])
        if (not self.in_paragraph
            and not self.in_title):
            self._append_text("\n")
    
    def visit_reference(self, node):
        tag = self._create_unique_tag()
        node.unique_tag = tag
        self._add_tag("a")
        self._add_tag(tag)
        
        def handle_click(event):
            get_workbench().open_url(node.attributes["refuri"])
            
        self.text.tag_bind(tag, "<ButtonRelease-1>", handle_click)
        
    def depart_reference(self, node):
        self._pop_tag("a")
        self._pop_tag(node.unique_tag)
    
    def visit_literal(self, node):
        self._add_tag("code")
    def depart_literal(self, node):
        self._pop_tag("code")
    
    def visit_bullet_list(self, node):
        self.active_lists.append(node.attributes["bullet"])
        
    def depart_bullet_list(self, node):
        self._append_text("\n")
        self.active_lists.pop()
    
    def visit_enumerated_list(self, node):
        self.active_lists.append(node.attributes["enumtype"])
        
    def depart_enumerated_list(self, node):
        self._append_text("\n")
        self.active_lists.pop()
    
    def visit_list_item(self, node):
        if self.active_lists[-1] == "*":
            self._append_text("• ")
        elif self.active_lists[-1] == "arabic":
            for i, sib in enumerate(node.parent.children):
                if sib is node:
                    self._append_text("%d. " % (i+1))
                    break

    def visit_note(self, node):
        self._add_tag("em")
        
    def depart_note(self, node):
        self._pop_tag("em")
        
    def visit_target(self, node):
        pass
    
    def visit_substitution_definition(self, node):
        raise docutils.nodes.SkipNode()
    
    def visit_system_message(self, node):
        logging.getLogger("thonny").warning("docutils message: '%s'. Context: %s" 
                                            % (node.astext(), node.parent))
        raise docutils.nodes.SkipNode
    
    def visit_emphasis(self, node):
        self._add_tag("em")    
    def depart_emphasis(self, node):
        self._pop_tag("em")    
    
    def visit_strong(self, node):
        self._add_tag("strong")    
    def depart_strong(self, node):
        self._pop_tag("strong")    
    
    def default_visit(self, node):
        self._append_text(self._node_to_text(node))
        print("skipping children", type(node), node)
        raise docutils.nodes.SkipChildren()

    def default_departure(self, node):
        # Pass all other nodes through.
        pass        
    
    def _create_unique_tag(self):
        self.unique_tag_count += 1
        return "_UT_%s" % self.unique_tag_count 
    
    def _node_to_text(self, node):
        return (node.astext()
                .replace("\r", "")
                .replace("\n", " "))
                          
    def _add_tag(self, tag):
        self._context_tags.append(tag)
        
    def _pop_tag(self, tag):
        self._context_tags.remove(tag)
    
    def _append_text(self, chars, extra_tags=()):
        #print("APPP", chars, tags)
        self.text.direct_insert("end", chars, self._get_effective_tags(extra_tags))
        
    def _append_image(self, name, extra_tags=()):
        index = self.text.index("end-1c")
        self.text.image_create(index, image=get_workbench().get_image(name))
        for tag in self._get_effective_tags(extra_tags):
            self.text.tag_add(tag, index)
            
    def _append_window(self, window, extra_tags=()):
        index = self.text.index("end-1c")
        self.text.window_create(index, window=window)
        for tag in self._get_effective_tags(extra_tags):
            self.text.tag_add(tag, index)
            
    def _get_effective_tags(self, extra_tags):
        tags = set(extra_tags) | set(self._context_tags)
        
        if self.active_lists:
            tags.add("list%d" % min(len(self.active_lists), 5))
        
        # combine tags
        if "code" in tags and "topic_title" in tags:
            tags.remove("code")
            tags.remove("topic_title")
            tags.add("topic_title_code")
        
        return tuple(sorted(tags))
        

def escape(s):
    return (s
            .replace("\\", "\\\\")
            .replace("*", "\\*")
            .replace("`", "\\`")
            .replace("_", "\\_")
            .replace("..", "\\.."))

def create_title(text, line_symbol="="):
    return (text + "\n" 
            + line_symbol * len(text) + "\n") 



from tkinter import ttk

from thonny.config_ui import ConfigurationPage
from thonny import get_workbench, ui_utils
from thonny.tktextext import TextFrame
from thonny.ui_utils import scrollbar_style


class AssistantConfigPage(ConfigurationPage):
    def __init__(self, master):
        super().__init__(master)
        
        self.add_checkbox("assistance.open_assistant_on_errors",
                          "Open Assistant automatically when program crashes with an exception",
                          row=1)
            
        self.add_checkbox("assistance.open_assistant_on_warnings",
                          "Open Assistant automatically when it has warnings for your code",
                          row=2)
        
        disabled_checks_label = ttk.Label(self, text="Disabled checks (one id per line)")
        disabled_checks_label.grid(row=3, sticky="nw", pady=(10,0))
        
        self.disabled_checks_box = TextFrame(self,vertical_scrollbar_style=scrollbar_style("Vertical"), 
                                            horizontal_scrollbar_style=scrollbar_style("Horizontal"),
                                            horizontal_scrollbar_class=ui_utils.AutoScrollbar,
                                            wrap="word",
                                            font="TkDefaultFont",
                                            #cursor="arrow",
                                            padx=5,
                                            pady=5,
                                            height=4,
                                            borderwidth=1,
                                            relief="groove"
                                            )
        self.disabled_checks_box.grid(row=4, sticky="nsew", pady=(0,10))
        self.disabled_checks_box.text.insert("1.0", "\n".join(
            get_workbench().get_option("assistance.disabled_checks")
        ))
        
        self.columnconfigure(0, weight=1)
        self.rowconfigure(4, weight=1)
    
    def apply(self):
        disabled_checks_str = (
            self.disabled_checks_box.text.get("1.0", "end")
            .replace("\r", "")
            .replace('"', "")
            .replace("'", "")
            .strip()
        )
        get_workbench().set_option("assistance.disabled_checks",
                                   disabled_checks_str.splitlines())


def load_plugin():
    get_workbench().add_configuration_page("Assistant", AssistantConfigPage)



class ADC:
    ''
    def read():
        pass

    def read_timed():
        pass

    def read_timed_multi():
        pass


class ADCAll:
    ''
    def read_channel():
        pass

    def read_core_temp():
        pass

    def read_core_vbat():
        pass

    def read_core_vref():
        pass

    def read_vref():
        pass


class Accel:
    ''
    def filtered_xyz():
        pass

    def read():
        pass

    def tilt():
        pass

    def write():
        pass

    def x():
        pass

    def y():
        pass

    def z():
        pass


class ExtInt:
    ''
    EVT_FALLING = 270663680
    EVT_RISING = 269615104
    EVT_RISING_FALLING = 271712256
    IRQ_FALLING = 270598144
    IRQ_RISING = 269549568
    IRQ_RISING_FALLING = 271646720
    def disable():
        pass

    def enable():
        pass

    def line():
        pass

    def regs():
        pass

    def swint():
        pass


class Flash:
    ''
    def ioctl():
        pass

    def readblocks():
        pass

    def writeblocks():
        pass


class I2C:
    ''
    MASTER = 0
    SLAVE = 1
    def deinit():
        pass

    def init():
        pass

    def is_ready():
        pass

    def mem_read():
        pass

    def mem_write():
        pass

    def recv():
        pass

    def scan():
        pass

    def send():
        pass


class LCD:
    ''
    def command():
        pass

    def contrast():
        pass

    def fill():
        pass

    def get():
        pass

    def light():
        pass

    def pixel():
        pass

    def show():
        pass

    def text():
        pass

    def write():
        pass


class LED:
    ''
    def intensity():
        pass

    def off():
        pass

    def on():
        pass

    def toggle():
        pass


class Pin:
    ''
    AF1_TIM1 = 1
    AF1_TIM2 = 1
    AF2_TIM3 = 2
    AF2_TIM4 = 2
    AF2_TIM5 = 2
    AF3_TIM10 = 3
    AF3_TIM11 = 3
    AF3_TIM9 = 3
    AF4_I2C1 = 4
    AF4_I2C3 = 4
    AF5_SPI1 = 5
    AF5_SPI2 = 5
    AF7_USART1 = 7
    AF7_USART2 = 7
    AF8_USART6 = 8
    AF9_I2C3 = 9
    AF_OD = 18
    AF_PP = 2
    ALT = 2
    ALT_OPEN_DRAIN = 18
    ANALOG = 3
    IN = 0
    IRQ_FALLING = 270598144
    IRQ_RISING = 269549568
    OPEN_DRAIN = 17
    OUT = 1
    OUT_OD = 17
    OUT_PP = 1
    PULL_DOWN = 2
    PULL_NONE = 0
    PULL_UP = 1
    def af():
        pass

    def af_list():
        pass

    board = None
    cpu = None
    def debug():
        pass

    def dict():
        pass

    def gpio():
        pass

    def high():
        pass

    def init():
        pass

    def irq():
        pass

    def low():
        pass

    def mapper():
        pass

    def mode():
        pass

    def name():
        pass

    def names():
        pass

    def off():
        pass

    def on():
        pass

    def pin():
        pass

    def port():
        pass

    def pull():
        pass

    def value():
        pass


class RTC:
    ''
    def calibration():
        pass

    def datetime():
        pass

    def info():
        pass

    def init():
        pass

    def wakeup():
        pass

SD = None

class SDCard:
    ''
    def info():
        pass

    def ioctl():
        pass

    def power():
        pass

    def present():
        pass

    def read():
        pass

    def readblocks():
        pass

    def write():
        pass

    def writeblocks():
        pass


class SPI:
    ''
    LSB = 128
    MASTER = 260
    MSB = 0
    SLAVE = 0
    def deinit():
        pass

    def init():
        pass

    def read():
        pass

    def readinto():
        pass

    def recv():
        pass

    def send():
        pass

    def send_recv():
        pass

    def write():
        pass

    def write_readinto():
        pass


class Servo:
    ''
    def angle():
        pass

    def calibration():
        pass

    def pulse_width():
        pass

    def speed():
        pass


class Switch:
    ''
    def callback():
        pass

    def value():
        pass


class Timer:
    ''
    BOTH = 10
    CENTER = 32
    DOWN = 16
    ENC_A = 9
    ENC_AB = 11
    ENC_B = 10
    FALLING = 2
    HIGH = 0
    IC = 8
    LOW = 2
    OC_ACTIVE = 3
    OC_FORCED_ACTIVE = 6
    OC_FORCED_INACTIVE = 7
    OC_INACTIVE = 4
    OC_TIMING = 2
    OC_TOGGLE = 5
    PWM = 0
    PWM_INVERTED = 1
    RISING = 0
    UP = 0
    def callback():
        pass

    def channel():
        pass

    def counter():
        pass

    def deinit():
        pass

    def freq():
        pass

    def init():
        pass

    def period():
        pass

    def prescaler():
        pass

    def source_freq():
        pass


class UART:
    ''
    CTS = 512
    RTS = 256
    def any():
        pass

    def deinit():
        pass

    def init():
        pass

    def read():
        pass

    def readchar():
        pass

    def readinto():
        pass

    def readline():
        pass

    def sendbreak():
        pass

    def write():
        pass

    def writechar():
        pass


class USB_HID:
    ''
    def recv():
        pass

    def send():
        pass


class USB_VCP:
    ''
    def any():
        pass

    def close():
        pass

    def isconnected():
        pass

    def read():
        pass

    def readinto():
        pass

    def readline():
        pass

    def readlines():
        pass

    def recv():
        pass

    def send():
        pass

    def setinterrupt():
        pass

    def write():
        pass

def bootloader():
    pass

def delay():
    pass

def dht_readinto():
    pass

def disable_irq():
    pass

def elapsed_micros():
    pass

def elapsed_millis():
    pass

def enable_irq():
    pass

def fault_debug():
    pass

def freq():
    pass

def hard_reset():
    pass

def have_cdc():
    pass

def hid():
    pass

hid_keyboard = None
hid_mouse = None
def info():
    pass

def main():
    pass

def micros():
    pass

def millis():
    pass

def mount():
    pass

def pwm():
    pass

def repl_info():
    pass

def repl_uart():
    pass

def servo():
    pass

def standby():
    pass

def stop():
    pass

def sync():
    pass

def udelay():
    pass

def unique_id():
    pass

def usb_mode():
    pass

def wfi():
    pass



from thonny import get_workbench
from thonny.workbench import SyntaxThemeSettings


def default_light() -> SyntaxThemeSettings:
    default_fg = "black"
    default_bg = "#fdfdfd"
    light_fg = "DarkGray"
    string_fg = "DarkGreen"
    open_string_bg = "#c3f9d3"
    
    return {
        "TEXT"          : {"foreground" : default_fg, "insertbackground" : default_fg,
                           "background" : default_bg},
        "GUTTER"        : {"foreground" : "#999999", "background" : "#e0e0e0"},
        "breakpoint"    : {"foreground" : "crimson"},
        "current_line"  : {"background" : "#f5f5f5"},
        
        "definition"    : {"foreground" : "DarkBlue", "font" : "BoldEditorFont"},
        "string"        : {"foreground" : string_fg},
        "string3"       : {"foreground" : string_fg},
        "open_string"   : {"foreground" : string_fg, "background" : open_string_bg},
        "open_string3"  : {"foreground" : string_fg, "background" : open_string_bg},
        "keyword"       : {"foreground" : "#7f0055", "font" : "BoldEditorFont"},
        "builtin"       : {"foreground" : "#7f0055"},
        "number"        : {"foreground" : "#B04600"},
        "comment"       : {"foreground" : light_fg},
        
        "prompt"        : {"foreground" : "purple", "font" : "BoldEditorFont"},
        "magic"         : {"foreground" : light_fg},
        "stdin"         : {"foreground" : "Blue"},
        "stdout"        : {"foreground" : "Black"},
        "stderr"        : {"foreground" : "Red"},
        "value"         : {"foreground" : "DarkBlue"},
        "hyperlink"     : {"foreground" : "#3A66DD", "underline" : True},
        
        # paren matcher
        "surrounding_parens"  : {"foreground" : "Blue", "font" : "BoldEditorFont"},
        "unclosed_expression"  : {"background" : "LightGray"},
        
        # find/replace
        "found" : {"foreground" : "blue", "underline" : True},
        "current_found" : {"foreground" : "white", "background" : "red"},
        
        "matched_name" : {"background" : "#e6ecfe"},
        "local_name" : {"font" : "ItalicEditorFont"},
         
        # debugger
        "active_focus"    : {"background" : "#F8FC9A", "borderwidth" : 1, "relief" : "solid"},
        "suspended_focus" : {"background" : ""       , "borderwidth" : 1, "relief" : "solid"},
        "completed_focus" : {"background" : "#BBEDB2", "borderwidth" : 1, "relief" : "flat"},
        "exception_focus" : {"background" : "#FFBFD6", "borderwidth" : 1, "relief" : "solid"},
        "expression_box"  : {"background" : "#DCEDF2", "foreground" : default_fg},
    }

def default_dark() -> SyntaxThemeSettings:
    default_fg = "#B3B3B3" 
    string_fg = "#8DC76F"
    open_string_bg = "#224533"
        
    #s.configure("Local.Code", foreground="#BCCAE8")
    #s.configure("MatchedName.Code", background="#193022")
    
    return {
        "TEXT"          : {"foreground" : default_fg, "insertbackground" : default_fg,
                           "background" : "#2d2d2d"},
        "GUTTER"        : {"foreground" : "#606060", "background" : "#323232"},
        "breakpoint"    : {"foreground" : "pink"},
        "current_line"  : {"background" : "#363636"},
        "sel"           : {"foreground" : "#eeeeee", "background" : "#6E6E6E"},
        
        "definition"    : {"foreground" : default_fg},
        "string"        : {"foreground" : string_fg},
        "string3"       : {"foreground" : string_fg},
        "open_string"   : {"foreground" : string_fg, "background" : open_string_bg},
        "open_string3"  : {"foreground" : string_fg, "background" : open_string_bg},
        "builtin"       : {"foreground" : "#A9B1C9"},
        "keyword"       : {"foreground" : "#A9B1C9", "font" : "BoldEditorFont"},
        "number"        : {"foreground" : "#FFCABF"},
        "comment"       : {"foreground" : "#D4D44E"},
        
        # shell
        "prompt"        : {"foreground" : "#5BEBBB", "font" : "BoldEditorFont"},
        "magic"         : {"foreground" : "pink"},
        "stdin"         : {"foreground" : "LightBlue"},
        "stdout"        : {"foreground" : "LightGray"},
        "stderr"        : {"foreground" : "#EB5B83"},
        "value"         : {"foreground" : "#EBEB5B"},
        "hyperlink"     : {"foreground" : "#DC5BEB", "underline" : True},
        
        # paren matcher
        "surrounding_parens"  : {"foreground" : "#F0995B", "font" : "BoldEditorFont"},
        "unclosed_expression"  : {"background" : "#000000"},
        
        # find/replace
        "found" : {"underline" : True},
        "current_found" : {"foreground" : "white", "background" : "red"},
        
        "matched_name" : {"background" : "#474747"}, 
        "local_name" : {"font" : "ItalicEditorFont"},
        
        # debugger
        "active_focus"    : {"background" : "#807238", "borderwidth" : 1, "relief" : "solid"},
        "suspended_focus" : {"background" : ""       , "borderwidth" : 1, "relief" : "solid"},
        "completed_focus" : {"background" : "#807238", "borderwidth" : 1, "relief" : "flat"},
        "exception_focus" : {"background" : "#FFBFD6", "borderwidth" : 1, "relief" : "solid"},
        "expression_box"  : {"background" : "#506E67", "foreground" : default_fg},
    }

def default_dark_green() -> SyntaxThemeSettings:
    open_string_bg = "#453B22"
    
    return {
        "TEXT"          : {"background" : "#273627"},
        "GUTTER"        : {"background" : "#33402F"},
        "current_line"  : {"background" : "#2E402E"},
        "sel"           : {"background" : "#6E6E6E"},
        "unclosed_expression"  : {"background" : "#0F1F15"},
        
        "open_string"   : {"background" : open_string_bg},
        "open_string3"  : {"background" : open_string_bg},

        "keyword"       : {"foreground" : "#88CFB6", "font" : "BoldEditorFont"},
        "builtin"       : {"foreground" : "#88CFB6"},
        
        # debugger
        "active_focus"    : {"background" : "#807238"},
        "completed_focus" : {"background" : "#807238"},
        "exception_focus" : {"background" : "#FFBFD6"},
        "expression_box"  : {"background" : "#506E67"},
    }
    

def default_dark_blue() -> SyntaxThemeSettings:
    open_string_bg = "#224533"
    return {
        "TEXT"          : {"background" : "#272936"},
        "GUTTER"        : {"background" : "#2F3640"},
        "current_line"  : {"background" : "#2D3040"},
        "sel"           : {"background" : "#6E6E6E"},
        "unclosed_expression"  : {"background" : "#100B21"},
        "open_string"   : {"background" : open_string_bg},
        "open_string3"  : {"background" : open_string_bg},
        
        "keyword"       : {"foreground" : "#8899CF", "font" : "BoldEditorFont"},
        "builtin"       : {"foreground" : "#8899CF"},

        # debugger
        "active_focus"    : {"background" : "#807238"},
        "completed_focus" : {"background" : "#807238"},
        "exception_focus" : {"background" : "#FFBFD6"},
        "expression_box"  : {"background" : "#506E67"},
    }
    

def idle_classic() -> SyntaxThemeSettings:
    string_fg = "#00aa00"
    return {
        "TEXT"          : {"foreground" : "black", "insertbackground" : "black", "background" : "white"},
        "GUTTER"        : {"foreground" : "gray", "background" : "#efefef"},
        "sel"           : {"foreground" : "black", "background" : "gray"},
        
        "number"        : {"foreground" : "black"},
        "definition"    : {"foreground" : "#0000ff", "font" : "EditorFont"},
        "string"        : {"foreground" : string_fg},
        "string3"       : {"foreground" : string_fg},
        "open_string"   : {"foreground" : string_fg},
        "open_string3"  : {"foreground" : string_fg},
        "keyword"       : {"foreground" : "#ff7700", "font" : "EditorFont"},
        "builtin"       : {"foreground" : "#900090"},
        "comment"       : {"foreground" : "#dd0000"},
        
        "prompt"        : {"foreground" : "#770000"},
        "stdin"         : {"foreground" : "black"},
        "stdout"        : {"foreground" : "Blue"},
        "value"         : {"foreground" : "Blue"},
        "stderr"        : {"foreground" : "Red"},
        
        "found"                 : {"foreground" : "", "underline" : True},
        "current_found"         : {"foreground" : "white", "background" : "black"},
    }
        

def idle_dark() -> SyntaxThemeSettings:
    normal_fg = "white"
    string_fg = "#02ff02"
    
    return {
        "TEXT"          : {"foreground" : normal_fg, "insertbackground" : normal_fg, "background" : "#002240"},
        "sel"           : {"foreground" : "#FFFFFF", "background" : "#7e7e7e"},
        
        "number"        : {"foreground" : normal_fg},
        "definition"    : {"foreground" : "#5e5eff", "font" : "EditorFont"},
        "string"        : {"foreground" : string_fg},
        "string3"       : {"foreground" : string_fg},
        "open_string"   : {"foreground" : string_fg},
        "open_string3"  : {"foreground" : string_fg},
        "keyword"       : {"foreground" : "#ff8000", "font" : "EditorFont"},
        "builtin"       : {"foreground" : "#ff00ff"},
        "comment"       : {"foreground" : "#dd0000"},
        
        "prompt"        : {"foreground" : "#ff4d4d"},
        "stdin"         : {"foreground" : normal_fg},
        "stdout"        : {"foreground" : "#c2d1fa"},
        "value"         : {"foreground" : "#c2d1fa"},
        "stderr"        : {"foreground" : "#ffb3b3"},
        
        "found"                 : {"foreground" : "", "underline" : True},
        "current_found"         : {"foreground" : "#002240", "background" : "#fbfbfb"},
    }

def desert_sunset() -> SyntaxThemeSettings:
    normal_fg = "#f0e68c"
    string_fg = "#ffa0a0"
    
    return {
        "TEXT"          : {"foreground" : normal_fg, "insertbackground" : normal_fg, 
                           "background" : "#333333"},
        "GUTTER"        : {"foreground" : "gray", "background" : "#404040"},
        "sel"           : {"foreground" : "#000000", "background" : "gray"},
        
        "number"        : {"foreground" : normal_fg},
        "definition"    : {"foreground" : "#98fb98"},
        "string"        : {"foreground" : string_fg},
        "string3"       : {"foreground" : string_fg},
        "open_string"   : {"foreground" : string_fg},
        "open_string3"  : {"foreground" : string_fg},
        "keyword"       : {"foreground" : "#cc6600"},
        "builtin"       : {"foreground" : "#519e51"},
        "comment"       : {"foreground" : "#87ceeb"},
        
        "prompt"        : {"foreground" : "#87ceeb"},
        "stdin"         : {"foreground" : normal_fg},
        "stdout"        : {"foreground" : "#eeeeee"},
        "value"         : {"foreground" : "#eeeeee"},
        "stderr"        : {"foreground" : "#ff3e40"},
        
        "found"                 : {"foreground" : "", "underline" : True},
        "current_found"         : {"foreground" : "#ffffff", "background" : "#333333"},
    }
    
def zenburn() -> SyntaxThemeSettings:
    # https://github.com/mig/gedit-themes/blob/master/zenburn.xml
    # https://github.com/trusktr/gedit-color-schemes/blob/master/gtksourceview-3.0/styles/zenburn.xml
    normal_fg = "#dcdccc"
    string_fg = "#cc9393"
    
    return {
        "TEXT"          : {"foreground" : normal_fg, "insertbackground" : normal_fg, 
                           "background" : "#3f3f3f"},
        "GUTTER"        : {"foreground" : "#7f8f8f", "background" : "#464646"},
        "current_line"  : {"background" : "#4A4A4A"},
        "sel"           : {"foreground" : "white", "background" : "#506070"},
        
        "number"        : {"foreground" : "#8cd0d3"},
        "definition"    : {"foreground" : "#f4a020", "font" : "BoldEditorFont"},
        "string"        : {"foreground" : string_fg},
        "string3"       : {"foreground" : string_fg},
        "open_string"   : {"foreground" : string_fg},
        "open_string3"  : {"foreground" : string_fg},
        "keyword"       : {"foreground" : "#f0dfaf", "font" : "BoldEditorFont"},
        "builtin"       : {"foreground" : "#efef8f"},
        "comment"       : {"foreground" : "#7f9f7f"},
        
        "prompt"        : {"foreground" : "#87ceeb"},
        "stdin"         : {"foreground" : normal_fg},
        "stdout"        : {"foreground" : "#eeeeee"},
        "value"         : {"foreground" : "#eeeeee"},
        "stderr"        : {"foreground" : "#ff3e40"},
        
        # paren matcher
        "surrounding_parens"  : {"foreground" : "white", "font" : "BoldEditorFont"},
    }
    
    
def load_plugin() -> None:
    get_workbench().add_syntax_theme("Default Light", None, default_light)
    get_workbench().add_syntax_theme("Default Dark", None, default_dark)
    get_workbench().add_syntax_theme("Default Dark Green", "Default Dark", default_dark_green)
    get_workbench().add_syntax_theme("Default Dark Blue", "Default Dark", default_dark_blue)
    get_workbench().add_syntax_theme("Desert Sunset", "Default Dark", desert_sunset)
    get_workbench().add_syntax_theme("Zenburn", "Default Dark", zenburn)
    get_workbench().add_syntax_theme("IDLE Classic", "Default Light", idle_classic)
    
    # Comments in IDLE Dark really hurt the eyes
    #get_workbench().add_syntax_theme("IDLE Dark", "Default Dark", idle_dark)
    
    get_workbench().set_default("view.syntax_theme", "Default Light")



# -*- coding: utf-8 -*-
 
import ast
import tkinter as tk
import traceback

from thonny import ast_utils, get_workbench, ui_utils
from thonny.common import TextRange, range_contains_smaller


class AstView(ui_utils.TreeFrame):
    def __init__(self, master):
        ui_utils.TreeFrame.__init__(self, master,
            columns=('range', 'lineno', 'col_offset', 'end_lineno', 'end_col_offset'),
            displaycolumns=(0,)
        )
        
        self._current_code_view = None
        self.tree.bind("<<TreeviewSelect>>", self._locate_code)
        self.tree.bind("<<Copy>>", self._copy_to_clipboard)
        get_workbench().get_editor_notebook().bind("<<NotebookTabChanged>>", self._update)
        get_workbench().bind("Save", self._update, True)
        get_workbench().bind("SaveAs", self._update, True)
        get_workbench().bind_class("Text", "<Double-Button-1>", self._update, True)

        self.tree.column('#0', width=550, anchor=tk.W)
        self.tree.column('range', width=100, anchor=tk.W)
        self.tree.column('lineno', width=30, anchor=tk.W)
        self.tree.column('col_offset', width=30, anchor=tk.W)
        self.tree.column('end_lineno', width=30, anchor=tk.W)
        self.tree.column('end_col_offset', width=30, anchor=tk.W)
        
        self.tree.heading('#0', text="Node", anchor=tk.W)
        self.tree.heading('range', text='Code range', anchor=tk.W)
        
        self.tree['show'] = ('headings', 'tree')
        self._current_source = None
        
        self._update(None)
    
    def _update(self, event):
        editor = get_workbench().get_editor_notebook().get_current_editor()
        
        if not editor:
            self._current_code_view = None
            return
        
        self._current_code_view = editor.get_code_view()
        self._current_source = self._current_code_view.get_content_as_bytes()
        selection = self._current_code_view.get_selected_range()

        self._clear_tree()
        
        if not self._current_source.strip():
            return
        
        try:
            root = ast_utils.parse_source(self._current_source)
            selected_ast_node = _find_closest_containing_node(root, selection)
            
        except Exception as e:
            self.tree.insert("", "end", text=str(e), open=True)
            traceback.print_exc()
            return
        
        def _format(key, node, parent_id):
            
            
            if isinstance(node, ast.AST):
                fields = [(key, val) for key, val in ast.iter_fields(node)]
                
                value_label = node.__class__.__name__
                    
            elif isinstance(node, list):
                fields = list(enumerate(node))
                if len(node) == 0:
                    value_label = "[]"
                else:
                    value_label = "[...]"
            else:
                fields = []
                value_label = repr(node)
            
            item_text = str(key) + "=" + value_label
            node_id = self.tree.insert(parent_id, "end", text=item_text, open=True)
            if node == selected_ast_node:
                self.tree.see(node_id)
                self.tree.selection_add(node_id)
            
            if hasattr(node, "lineno") and hasattr(node, "col_offset"):
                self.tree.set(node_id, "lineno", node.lineno)
                self.tree.set(node_id, "col_offset", node.col_offset)
                
                range_str = str(node.lineno) + '.' + str(node.col_offset)
                if hasattr(node, "end_lineno") and hasattr(node, "end_col_offset"):
                    self.tree.set(node_id, "end_lineno", node.end_lineno)
                    self.tree.set(node_id, "end_col_offset", node.end_col_offset)
                    range_str += "  -  " + str(node.end_lineno) + '.' + str(node.end_col_offset)
                else:
                    # fallback
                    self.tree.set(node_id, "end_lineno", node.lineno)
                    self.tree.set(node_id, "end_col_offset", node.col_offset + 1)
                    
                self.tree.set(node_id, "range", range_str)
                
            for field_key, field_value in fields:
                _format(field_key, field_value, node_id)
                
                
        _format("root", root, "")
    
    def _locate_code(self, event):
        if self._current_code_view is None:
            return
        
        iid = self.tree.focus()
        
        if iid != '':
            values = self.tree.item(iid)['values']
            if isinstance(values, list) and len(values) >= 5:
                start_line, start_col, end_line, end_col = values[1:5] 
                self._current_code_view.select_range(TextRange(start_line, start_col, 
                                                    end_line, end_col))
    
    def _clear_tree(self):
        for child_id in self.tree.get_children():
            self.tree.delete(child_id)
    
    def _copy_to_clipboard(self, event):
        self.clipboard_clear()
        if self._current_source is not None:
            pretty_ast = ast_utils.pretty(ast_utils.parse_source(self._current_source))
            self.clipboard_append(pretty_ast)

def _find_closest_containing_node(tree, text_range):
    # first look among children
    for child in ast.iter_child_nodes(tree):
        result = _find_closest_containing_node(child, text_range)
        if result is not None:
            return result

    # no suitable child was found
    if (hasattr(tree, "lineno")
        and range_contains_smaller( 
            TextRange(tree.lineno, tree.col_offset, tree.end_lineno, tree.end_col_offset),
            text_range
        )):
        return tree
    # nope
    else:
        return None

            
def load_plugin() -> None: 
    get_workbench().add_view(AstView, "AST", "s")



import logging
import tkinter as tk
from tkinter import ttk

from thonny import get_workbench
from thonny.config_ui import ConfigurationPage


class EditorConfigurationPage(ConfigurationPage):
    
    def __init__(self, master):
        ConfigurationPage.__init__(self, master)
        
        try:
            self.add_checkbox("view.name_highlighting", "Highlight matching names")
        except Exception:
            # name matcher may have been disabled
            logging.warning("Couldn't create name matcher checkbox")
            
        try:
            self.add_checkbox("view.locals_highlighting", "Highlight local variables")
        except Exception:
            # locals highlighter may have been disabled
            logging.warning("Couldn't create name locals highlighter checkbox")
            
        self.add_checkbox("view.paren_highlighting", "Highlight parentheses")
        self.add_checkbox("view.syntax_coloring", "Highlight syntax elements")
        self.add_checkbox("view.highlight_current_line", "Highlight current line (requires reopening the editor)")
        
        self.add_checkbox("edit.tab_complete_in_editor", "Allow code completion with Tab-key in editors", 
                          columnspan=2, pady=(20,0), )
        self.add_checkbox("edit.tab_complete_in_shell",  "Allow code completion with Tab-key in Shell",
                          columnspan=2)
        
        self.add_checkbox("view.show_line_numbers", "Show line numbers", pady=(20,0))
        self._line_length_var = get_workbench().get_variable("view.recommended_line_length")
        label = ttk.Label(self, text="Recommended maximum line length\n(Set to 0 to turn off margin line)")
        label.grid(row=20, column=0, sticky=tk.W)
        self._line_length_combo = ttk.Combobox(self, width=4,
                                        exportselection=False,
                                        textvariable=self._line_length_var,
                                        state='readonly',
                                        values=[0,60,70,80,90,100,110,120])
        self._line_length_combo.grid(row=20, column=1, sticky=tk.E)
        
        self.columnconfigure(0, weight=1)
    
    def apply(self):
        ConfigurationPage.apply(self)
        get_workbench().get_editor_notebook().update_appearance()
    

def load_plugin() -> None:
    get_workbench().add_configuration_page("Editor", EditorConfigurationPage)



def cos():
    pass

e = 2.718282
def exp():
    pass

def log():
    pass

def log10():
    pass

def phase():
    pass

pi = 3.141593
def polar():
    pass

def rect():
    pass

def sin():
    pass

def sqrt():
    pass



"""Facilities for learning the structure of incomplete Python code
Mostly copied/adapted from idlelib.HyperParser and idlelib.PyParse
"""



import re
import string
from collections import Mapping
from keyword import iskeyword
from typing import Dict  # @UnusedImport

NUM_CONTEXT_LINES = (50, 500, 5000000)

# Reason last stmt is continued (or C_NONE if it's not).
(C_NONE, C_BACKSLASH, C_STRING_FIRST_LINE,
 C_STRING_NEXT_LINES, C_BRACKET) = range(5)

# Find what looks like the start of a popular stmt.

_synchre = re.compile(r"""
    ^
    [ \t]*
    (?: while
    |   else
    |   def
    |   return
    |   assert
    |   break
    |   class
    |   continue
    |   elif
    |   try
    |   except
    |   raise
    |   import
    |   yield
    )
    \b
""", re.VERBOSE | re.MULTILINE).search

# Match blank line or non-indenting comment line.

_junkre = re.compile(r"""
    [ \t]*
    (?: \# \S .* )?
    \n
""", re.VERBOSE).match

# Match any flavor of string; the terminating quote is optional
# so that we're robust in the face of incomplete program text.

_match_stringre = re.compile(r"""
    \""" [^"\\]* (?:
                     (?: \\. | "(?!"") )
                     [^"\\]*
                 )*
    (?: \""" )?

|   " [^"\\\n]* (?: \\. [^"\\\n]* )* "?

|   ''' [^'\\]* (?:
                   (?: \\. | '(?!'') )
                   [^'\\]*
                )*
    (?: ''' )?

|   ' [^'\\\n]* (?: \\. [^'\\\n]* )* '?
""", re.VERBOSE | re.DOTALL).match

# Match a line that starts with something interesting;
# used to find the first item of a bracket structure.

_itemre = re.compile(r"""
    [ \t]*
    [^\s#\\]    # if we match, m.end()-1 is the interesting char
""", re.VERBOSE).match

# Match start of stmts that should be followed by a dedent.

_closere = re.compile(r"""
    \s*
    (?: return
    |   break
    |   continue
    |   raise
    |   pass
    )
    \b
""", re.VERBOSE).match

# Chew up non-special chars as quickly as possible.  If match is
# successful, m.end() less 1 is the index of the last boring char
# matched.  If match is unsuccessful, the string starts with an
# interesting char.

_chew_ordinaryre = re.compile(r"""
    [^[\](){}#'"\\]+
""", re.VERBOSE).match


class StringTranslatePseudoMapping(Mapping):
    r"""Utility class to be used with str.translate()

    This Mapping class wraps a given dict. When a value for a key is
    requested via __getitem__() or get(), the key is looked up in the
    given dict. If found there, the value from the dict is returned.
    Otherwise, the default value given upon initialization is returned.

    This allows using str.translate() to make some replacements, and to
    replace all characters for which no replacement was specified with
    a given character instead of leaving them as-is.

    For example, to replace everything except whitespace with 'x':

    >>> whitespace_chars = ' \t\n\r'
    >>> preserve_dict = {ord(c): ord(c) for c in whitespace_chars}
    >>> mapping = StringTranslatePseudoMapping(preserve_dict, ord('x'))
    >>> text = "a + b\tc\nd"
    >>> text.translate(mapping)
    'x x x\tx\nx'
    """
    def __init__(self, non_defaults, default_value):
        self._non_defaults = non_defaults
        self._default_value = default_value

        def _get(key, _get=non_defaults.get, _default=default_value):
            return _get(key, _default)
        self._get = _get

    def __getitem__(self, item):
        return self._get(item)

    def __len__(self):
        return len(self._non_defaults)

    def __iter__(self):
        return iter(self._non_defaults)

    def get(self, key, default=None):
        return self._get(key)


class RoughParser:

    def __init__(self, indentwidth, tabwidth):
        self.indentwidth = indentwidth
        self.tabwidth = tabwidth

    def set_str(self, s):
        assert len(s) == 0 or s[-1] == '\n'
        self.str = s
        self.study_level = 0

    # Return index of a good place to begin parsing, as close to the
    # end of the string as possible.  This will be the start of some
    # popular stmt like "if" or "def".  Return None if none found:
    # the caller should pass more prior context then, if possible, or
    # if not (the entire program text up until the point of interest
    # has already been tried) pass 0 to set_lo.
    #
    # This will be reliable iff given a reliable is_char_in_string
    # function, meaning that when it says "no", it's absolutely
    # guaranteed that the char is not in a string.

    def find_good_parse_start(self, is_char_in_string=None,
                              _synchre=_synchre):
        # pylint: disable=redefined-builtin
        
        str, pos = self.str, None  # @ReservedAssignment

        if not is_char_in_string:
            # no clue -- make the caller pass everything
            return None

        # Peek back from the end for a good place to start,
        # but don't try too often; pos will be left None, or
        # bumped to a legitimate synch point.
        limit = len(str)
        for _ in range(5): 
            i = str.rfind(":\n", 0, limit)
            if i < 0:
                break
            i = str.rfind('\n', 0, i) + 1  # start of colon line
            m = _synchre(str, i, limit)
            if m and not is_char_in_string(m.start()):
                pos = m.start()
                break
            limit = i
        if pos is None:
            # Nothing looks like a block-opener, or stuff does
            # but is_char_in_string keeps returning true; most likely
            # we're in or near a giant string, the colorizer hasn't
            # caught up enough to be helpful, or there simply *aren't*
            # any interesting stmts.  In any of these cases we're
            # going to have to parse the whole thing to be sure, so
            # give it one last try from the start, but stop wasting
            # time here regardless of the outcome.
            m = _synchre(str)
            if m and not is_char_in_string(m.start()):
                pos = m.start()
            return pos

        # Peeking back worked; look forward until _synchre no longer
        # matches.
        i = pos + 1
        while 1:
            m = _synchre(str, i)
            if m:
                s, i = m.span()
                if not is_char_in_string(s):
                    pos = s
            else:
                break
        return pos

    # Throw away the start of the string.  Intended to be called with
    # find_good_parse_start's result.

    def set_lo(self, lo):
        assert lo == 0 or self.str[lo-1] == '\n'
        if lo > 0:
            self.str = self.str[lo:]

    # Build a translation table to map uninteresting chars to 'x', open
    # brackets to '(', close brackets to ')' while preserving quotes,
    # backslashes, newlines and hashes. This is to be passed to
    # str.translate() in _study1().
    _tran1 = {} # type: Dict[int, int]
    _tran1.update((ord(c), ord('(')) for c in "({[")
    _tran1.update((ord(c), ord(')')) for c in ")}]")
    _tran1.update((ord(c), ord(c)) for c in "\"'\\\n#")
    _tran = StringTranslatePseudoMapping(_tran1, default_value=ord('x'))

    # As quickly as humanly possible <wink>, find the line numbers (0-
    # based) of the non-continuation lines.
    # Creates self.{goodlines, continuation}.

    def _study1(self):
        # pylint: disable=redefined-builtin

        if self.study_level >= 1:
            return
        self.study_level = 1

        # Map all uninteresting characters to "x", all open brackets
        # to "(", all close brackets to ")", then collapse runs of
        # uninteresting characters.  This can cut the number of chars
        # by a factor of 10-40, and so greatly speed the following loop.
        str = (self.str  # @ReservedAssignment
                .translate(self._tran)
                .replace('xxxxxxxx', 'x')
                .replace('xxxx', 'x')
                .replace('xx', 'x')
                .replace('xx', 'x')
                .replace('\nx', '\n'))
        # note that replacing x\n with \n would be incorrect, because
        # x may be preceded by a backslash

        # March over the squashed version of the program, accumulating
        # the line numbers of non-continued stmts, and determining
        # whether & why the last stmt is a continuation.
        continuation = C_NONE
        level = lno = 0     # level is nesting level; lno is line number
        self.goodlines = goodlines = [0]
        push_good = goodlines.append
        i, n = 0, len(str)
        while i < n:
            ch = str[i]
            i = i+1

            # cases are checked in decreasing order of frequency
            if ch == 'x':
                continue

            if ch == '\n':
                lno = lno + 1
                if level == 0:
                    push_good(lno)
                    # else we're in an unclosed bracket structure
                continue

            if ch == '(':
                level = level + 1
                continue

            if ch == ')':
                if level:
                    level = level - 1
                    # else the program is invalid, but we can't complain
                continue

            if ch == '"' or ch == "'":
                # consume the string
                quote = ch
                if str[i-1:i+2] == quote * 3:
                    quote = quote * 3
                firstlno = lno
                w = len(quote) - 1
                i = i+w
                while i < n:
                    ch = str[i]
                    i = i+1

                    if ch == 'x':
                        continue

                    if str[i-1:i+w] == quote:
                        i = i+w
                        break

                    if ch == '\n':
                        lno = lno + 1
                        if w == 0:
                            # unterminated single-quoted string
                            if level == 0:
                                push_good(lno)
                            break
                        continue

                    if ch == '\\':
                        assert i < n
                        if str[i] == '\n':
                            lno = lno + 1
                        i = i+1
                        continue

                    # else comment char or paren inside string

                else:
                    # didn't break out of the loop, so we're still
                    # inside a string
                    if (lno - 1) == firstlno:
                        # before the previous \n in str, we were in the first
                        # line of the string
                        continuation = C_STRING_FIRST_LINE
                    else:
                        continuation = C_STRING_NEXT_LINES
                continue    # with outer loop

            if ch == '#':
                # consume the comment
                i = str.find('\n', i)
                assert i >= 0
                continue

            assert ch == '\\'
            assert i < n
            if str[i] == '\n':
                lno = lno + 1
                if i+1 == n:
                    continuation = C_BACKSLASH
            i = i+1

        # The last stmt may be continued for all 3 reasons.
        # String continuation takes precedence over bracket
        # continuation, which beats backslash continuation.
        if (continuation != C_STRING_FIRST_LINE
            and continuation != C_STRING_NEXT_LINES and level > 0):
            continuation = C_BRACKET
        self.continuation = continuation

        # Push the final line number as a sentinel value, regardless of
        # whether it's continued.
        assert (continuation == C_NONE) == (goodlines[-1] == lno)
        if goodlines[-1] != lno:
            push_good(lno)

    def get_continuation_type(self):
        self._study1()
        return self.continuation

    # study1 was sufficient to determine the continuation status,
    # but doing more requires looking at every character.  study2
    # does this for the last interesting statement in the block.
    # Creates:
    #     self.stmt_start, stmt_end
    #         slice indices of last interesting stmt
    #     self.stmt_bracketing
    #         the bracketing structure of the last interesting stmt;
    #         for example, for the statement "say(boo) or die", stmt_bracketing
    #         will be [(0, 0), (3, 1), (8, 0)]. Strings and comments are
    #         treated as brackets, for the matter.
    #     self.lastch
    #         last non-whitespace character before optional trailing
    #         comment
    #     self.lastopenbracketpos
    #         if continuation is C_BRACKET, index of last open bracket

    def _study2(self):
        # pylint: disable=redefined-builtin

        if self.study_level >= 2:
            return
        self._study1()
        self.study_level = 2

        # Set p and q to slice indices of last interesting stmt.
        str, goodlines = self.str, self.goodlines  # @ReservedAssignment
        i = len(goodlines) - 1
        p = len(str)    # index of newest line
        while i:
            assert p
            # p is the index of the stmt at line number goodlines[i].
            # Move p back to the stmt at line number goodlines[i-1].
            q = p
            for _ in range(goodlines[i-1], goodlines[i]):  # @UnusedVariable
                # tricky: sets p to 0 if no preceding newline
                p = str.rfind('\n', 0, p-1) + 1
            # The stmt str[p:q] isn't a continuation, but may be blank
            # or a non-indenting comment line.
            if  _junkre(str, p):
                i = i-1
            else:
                break
        if i == 0:
            # nothing but junk!
            assert p == 0
            q = p
        self.stmt_start, self.stmt_end = p, q

        # Analyze this stmt, to find the last open bracket (if any)
        # and last interesting character (if any).
        lastch = ""
        stack = []  # stack of open bracket indices
        push_stack = stack.append
        bracketing = [(p, 0)]
        while p < q:
            # suck up all except ()[]{}'"#\\
            m = _chew_ordinaryre(str, p, q)
            if m:
                # we skipped at least one boring char
                newp = m.end()
                # back up over totally boring whitespace
                i = newp - 1    # index of last boring char
                while i >= p and str[i] in " \t\n":
                    i = i-1
                if i >= p:
                    lastch = str[i]
                p = newp
                if p >= q:
                    break

            ch = str[p]

            if ch in "([{":
                push_stack(p)
                bracketing.append((p, len(stack)))
                lastch = ch
                p = p+1
                continue

            if ch in ")]}":
                if stack:
                    del stack[-1]
                lastch = ch
                p = p+1
                bracketing.append((p, len(stack)))
                continue

            if ch == '"' or ch == "'":
                # consume string
                # Note that study1 did this with a Python loop, but
                # we use a regexp here; the reason is speed in both
                # cases; the string may be huge, but study1 pre-squashed
                # strings to a couple of characters per line.  study1
                # also needed to keep track of newlines, and we don't
                # have to.
                bracketing.append((p, len(stack)+1))
                lastch = ch
                p = _match_stringre(str, p, q).end()
                bracketing.append((p, len(stack)))
                continue

            if ch == '#':
                # consume comment and trailing newline
                bracketing.append((p, len(stack)+1))
                p = str.find('\n', p, q) + 1
                assert p > 0
                bracketing.append((p, len(stack)))
                continue

            assert ch == '\\'
            p = p+1     # beyond backslash
            assert p < q
            if str[p] != '\n':
                # the program is invalid, but can't complain
                lastch = ch + str[p]
            p = p+1     # beyond escaped char

        # end while p < q:

        self.lastch = lastch
        if stack:
            self.lastopenbracketpos = stack[-1]
        self.stmt_bracketing = tuple(bracketing)

    # Assuming continuation is C_BRACKET, return the number
    # of spaces the next line should be indented.

    def compute_bracket_indent(self):
        # pylint: disable=redefined-builtin
        self._study2()
        assert self.continuation == C_BRACKET
        j = self.lastopenbracketpos
        str = self.str  # @ReservedAssignment
        n = len(str)
        origi = i = str.rfind('\n', 0, j) + 1
        j = j+1     # one beyond open bracket
        # find first list item; set i to start of its line
        while j < n:
            m = _itemre(str, j)
            if m:
                j = m.end() - 1     # index of first interesting char
                extra = 0
                break
            else:
                # this line is junk; advance to next line
                i = j = str.find('\n', j) + 1
        else:
            # nothing interesting follows the bracket;
            # reproduce the bracket line's indentation + a level
            j = i = origi
            while str[j] in " \t":
                j = j+1
            extra = self.indentwidth
        return len(str[i:j].expandtabs(self.tabwidth)) + extra

    # Return number of physical lines in last stmt (whether or not
    # it's an interesting stmt!  this is intended to be called when
    # continuation is C_BACKSLASH).

    def get_num_lines_in_stmt(self):
        self._study1()
        goodlines = self.goodlines
        return goodlines[-1] - goodlines[-2]

    # Assuming continuation is C_BACKSLASH, return the number of spaces
    # the next line should be indented.  Also assuming the new line is
    # the first one following the initial line of the stmt.

    def compute_backslash_indent(self):
        # pylint: disable=redefined-builtin
        self._study2()
        assert self.continuation == C_BACKSLASH
        str = self.str  # @ReservedAssignment
        i = self.stmt_start
        while str[i] in " \t":
            i = i+1
        startpos = i

        # See whether the initial line starts an assignment stmt; i.e.,
        # look for an = operator
        endpos = str.find('\n', startpos) + 1
        found = level = 0
        while i < endpos:
            ch = str[i]
            if ch in "([{":
                level = level + 1
                i = i+1
            elif ch in ")]}":
                if level:
                    level = level - 1
                i = i+1
            elif ch == '"' or ch == "'":
                i = _match_stringre(str, i, endpos).end()
            elif ch == '#':
                break
            elif level == 0 and ch == '=' and \
                   (i == 0 or str[i-1] not in "=<>!") and \
                   str[i+1] != '=':
                found = 1
                break
            else:
                i = i+1

        if found:
            # found a legit =, but it may be the last interesting
            # thing on the line
            i = i+1     # move beyond the =
            found = re.match(r"\s*\\", str[i:endpos]) is None

        if not found:
            # oh well ... settle for moving beyond the first chunk
            # of non-whitespace chars
            i = startpos
            while str[i] not in " \t\n":
                i = i+1

        return len(str[self.stmt_start:i].expandtabs(\
                                     self.tabwidth)) + 1

    # Return the leading whitespace on the initial line of the last
    # interesting stmt.

    def get_base_indent_string(self):
        self._study2()
        i, n = self.stmt_start, self.stmt_end
        j = i
        str_ = self.str
        while j < n and str_[j] in " \t":
            j = j + 1
        return str_[i:j]

    # Did the last interesting stmt open a block?

    def is_block_opener(self):
        self._study2()
        return self.lastch == ':'

    # Did the last interesting stmt close a block?

    def is_block_closer(self):
        self._study2()
        return _closere(self.str, self.stmt_start) is not None

    # index of last open bracket ({[, or None if none
    lastopenbracketpos = None

    def get_last_open_bracket_pos(self):
        self._study2()
        return self.lastopenbracketpos

    # the structure of the bracketing of the last interesting statement,
    # in the format defined in _study2, or None if the text didn't contain
    # anything
    stmt_bracketing = None

    def get_last_stmt_bracketing(self):
        self._study2()
        return self.stmt_bracketing




# all ASCII chars that may be in an identifier
_ASCII_ID_CHARS = frozenset(string.ascii_letters + string.digits + "_")
# all ASCII chars that may be the first char of an identifier
_ASCII_ID_FIRST_CHARS = frozenset(string.ascii_letters + "_")

# lookup table for whether 7-bit ASCII chars are valid in a Python identifier
_IS_ASCII_ID_CHAR = [(chr(x) in _ASCII_ID_CHARS) for x in range(128)]
# lookup table for whether 7-bit ASCII chars are valid as the first
# char in a Python identifier
_IS_ASCII_ID_FIRST_CHAR = \
    [(chr(x) in _ASCII_ID_FIRST_CHARS) for x in range(128)]


class HyperParser:
    """Provide advanced parsing abilities for ParenMatch and other extensions.
    
    HyperParser uses PyParser.  PyParser mostly gives information on the
    proper indentation of code.  HyperParser gives additional information on
    the structure of code.
    """
    def __init__(self, text, index):
        "To initialize, analyze the surroundings of the given index."

        self.text = text

        parser = RoughParser(text.indentwidth, text.tabwidth)

        def index2line(index):
            return int(float(index))
        lno = index2line(text.index(index))

        for context in NUM_CONTEXT_LINES:
            startat = max(lno - context, 1)
            startatindex = repr(startat) + ".0"
            stopatindex = "%d.end" % lno
            # We add the newline because PyParse requires a newline
            # at end. We add a space so that index won't be at end
            # of line, so that its status will be the same as the
            # char before it, if should.
            parser.set_str(text.get(startatindex, stopatindex)+' \n')
            bod = parser.find_good_parse_start(
                      _build_char_in_string_func(startatindex))
            if bod is not None or startat == 1:
                break
        parser.set_lo(bod or 0)

        # We want what the parser has, minus the last newline and space.
        self.rawtext = parser.str[:-2]
        # Parser.str apparently preserves the statement we are in, so
        # that stopatindex can be used to synchronize the string with
        # the text box indices.
        self.stopatindex = stopatindex
        self.bracketing = parser.get_last_stmt_bracketing()
        # find which pairs of bracketing are openers. These always
        # correspond to a character of rawtext.
        self.isopener = [i>0 and self.bracketing[i][1] >
                         self.bracketing[i-1][1]
                         for i in range(len(self.bracketing))]

        self.set_index(index)

    def set_index(self, index):
        """Set the index to which the functions relate.

        The index must be in the same statement.
        """
        indexinrawtext = (len(self.rawtext) -
                          len(self.text.get(index, self.stopatindex)))
        if indexinrawtext < 0:
            raise ValueError("Index %s precedes the analyzed statement"
                             % index)
        self.indexinrawtext = indexinrawtext
        # find the rightmost bracket to which index belongs
        self.indexbracket = 0
        while (self.indexbracket < len(self.bracketing)-1 and
               self.bracketing[self.indexbracket+1][0] < self.indexinrawtext):
            self.indexbracket += 1
        if (self.indexbracket < len(self.bracketing)-1 and
            self.bracketing[self.indexbracket+1][0] == self.indexinrawtext and
           not self.isopener[self.indexbracket+1]):
            self.indexbracket += 1

    def is_in_string(self):
        """Is the index given to the HyperParser in a string?"""
        # The bracket to which we belong should be an opener.
        # If it's an opener, it has to have a character.
        return (self.isopener[self.indexbracket] and
                self.rawtext[self.bracketing[self.indexbracket][0]]
                in ('"', "'"))

    def is_in_code(self):
        """Is the index given to the HyperParser in normal code?"""
        return (not self.isopener[self.indexbracket] or
                self.rawtext[self.bracketing[self.indexbracket][0]]
                not in ('#', '"', "'"))

    def get_surrounding_brackets(self, openers='([{', mustclose=False):
        """Return bracket indexes or None.

        If the index given to the HyperParser is surrounded by a
        bracket defined in openers (or at least has one before it),
        return the indices of the opening bracket and the closing
        bracket (or the end of line, whichever comes first).

        If it is not surrounded by brackets, or the end of line comes
        before the closing bracket and mustclose is True, returns None.
        """

        bracketinglevel = self.bracketing[self.indexbracket][1]
        before = self.indexbracket
        while (not self.isopener[before] or
              self.rawtext[self.bracketing[before][0]] not in openers or
              self.bracketing[before][1] > bracketinglevel):
            before -= 1
            if before < 0:
                return None
            bracketinglevel = min(bracketinglevel, self.bracketing[before][1])
        after = self.indexbracket + 1
        while (after < len(self.bracketing) and
              self.bracketing[after][1] >= bracketinglevel):
            after += 1

        beforeindex = self.text.index("%s-%dc" %
            (self.stopatindex, len(self.rawtext)-self.bracketing[before][0]))
        if (after >= len(self.bracketing) or
           self.bracketing[after][0] > len(self.rawtext)):
            if mustclose:
                return None
            afterindex = self.stopatindex
        else:
            # We are after a real char, so it is a ')' and we give the
            # index before it.
            afterindex = self.text.index(
                "%s-%dc" % (self.stopatindex,
                 len(self.rawtext)-(self.bracketing[after][0]-1)))

        return beforeindex, afterindex

    # the set of built-in identifiers which are also keywords,
    # i.e. keyword.iskeyword() returns True for them
    _ID_KEYWORDS = frozenset({"True", "False", "None"})

    @classmethod
    def _eat_identifier(cls, s, limit, pos):
        """Given a string and pos, return the number of chars in the
        identifier which ends at pos, or 0 if there is no such one.

        This ignores non-identifier eywords are not identifiers.
        """
        is_ascii_id_char = _IS_ASCII_ID_CHAR

        # Start at the end (pos) and work backwards.
        i = pos

        # Go backwards as long as the characters are valid ASCII
        # identifier characters. This is an optimization, since it
        # is faster in the common case where most of the characters
        # are ASCII.
        while i > limit and (
                ord(s[i - 1]) < 128 and
                is_ascii_id_char[ord(s[i - 1])]
        ):
            i -= 1

        # If the above loop ended due to reaching a non-ASCII
        # character, continue going backwards using the most generic
        # test for whether a string contains only valid identifier
        # characters.
        if i > limit and ord(s[i - 1]) >= 128:
            while i - 4 >= limit and ('a' + s[i - 4:pos]).isidentifier():
                i -= 4
            if i - 2 >= limit and ('a' + s[i - 2:pos]).isidentifier():
                i -= 2
            if i - 1 >= limit and ('a' + s[i - 1:pos]).isidentifier():
                i -= 1

            # The identifier candidate starts here. If it isn't a valid
            # identifier, don't eat anything. At this point that is only
            # possible if the first character isn't a valid first
            # character for an identifier.
            if not s[i:pos].isidentifier():
                return 0
        elif i < pos:
            # All characters in str[i:pos] are valid ASCII identifier
            # characters, so it is enough to check that the first is
            # valid as the first character of an identifier.
            if not _IS_ASCII_ID_FIRST_CHAR[ord(s[i])]:
                return 0

        # All keywords are valid identifiers, but should not be
        # considered identifiers here, except for True, False and None.
        if i < pos and (
                iskeyword(s[i:pos]) and
                s[i:pos] not in cls._ID_KEYWORDS
        ):
            return 0

        return pos - i

    # This string includes all chars that may be in a white space
    _whitespace_chars = " \t\n\\"

    def get_expression(self):
        """Return a string with the Python expression which ends at the
        given index, which is empty if there is no real one.
        """
        if not self.is_in_code():
            raise ValueError("get_expression should only be called"
                             "if index is inside a code.")

        rawtext = self.rawtext
        bracketing = self.bracketing

        brck_index = self.indexbracket
        brck_limit = bracketing[brck_index][0]
        pos = self.indexinrawtext

        last_identifier_pos = pos
        postdot_phase = True

        while 1:
            # Eat whitespaces, comments, and if postdot_phase is False - a dot
            while 1:
                if pos>brck_limit and rawtext[pos-1] in self._whitespace_chars:
                    # Eat a whitespace
                    pos -= 1
                elif (not postdot_phase and
                      pos > brck_limit and rawtext[pos-1] == '.'):
                    # Eat a dot
                    pos -= 1
                    postdot_phase = True
                # The next line will fail if we are *inside* a comment,
                # but we shouldn't be.
                elif (pos == brck_limit and brck_index > 0 and
                      rawtext[bracketing[brck_index-1][0]] == '#'):
                    # Eat a comment
                    brck_index -= 2
                    brck_limit = bracketing[brck_index][0]
                    pos = bracketing[brck_index+1][0]
                else:
                    # If we didn't eat anything, quit.
                    break

            if not postdot_phase:
                # We didn't find a dot, so the expression end at the
                # last identifier pos.
                break

            ret = self._eat_identifier(rawtext, brck_limit, pos)
            if ret:
                # There is an identifier to eat
                pos = pos - ret
                last_identifier_pos = pos
                # Now, to continue the search, we must find a dot.
                postdot_phase = False
                # (the loop continues now)

            elif pos == brck_limit:
                # We are at a bracketing limit. If it is a closing
                # bracket, eat the bracket, otherwise, stop the search.
                level = bracketing[brck_index][1]
                while brck_index > 0 and bracketing[brck_index-1][1] > level:
                    brck_index -= 1
                if bracketing[brck_index][0] == brck_limit:
                    # We were not at the end of a closing bracket
                    break
                pos = bracketing[brck_index][0]
                brck_index -= 1
                brck_limit = bracketing[brck_index][0]
                last_identifier_pos = pos
                if rawtext[pos] in "([":
                    # [] and () may be used after an identifier, so we
                    # continue. postdot_phase is True, so we don't allow a dot.
                    pass
                else:
                    # We can't continue after other types of brackets
                    if rawtext[pos] in "'\"":
                        # Scan a string prefix
                        while pos > 0 and rawtext[pos - 1] in "rRbBuU":
                            pos -= 1
                        last_identifier_pos = pos
                    break

            else:
                # We've found an operator or something.
                break

        return rawtext[last_identifier_pos:self.indexinrawtext]



def _is_char_in_string(text_index):
    # in idlelib.EditorWindow this used info from colorer
    # to speed up things, but I dont want to rely on this
    return 1
    
def _build_char_in_string_func(startindex):
    # copied from idlelib.EditorWindow (Python 3.4.2)
    
    # Our editwin provides a _is_char_in_string function that works
    # with a Tk text index, but PyParse only knows about offsets into
    # a string. This builds a function for PyParse that accepts an
    # offset.

    def inner(offset, _startindex=startindex,
              _icis=_is_char_in_string):
        return _icis(_startindex + "+%dc" % offset)
    return inner



import re
import subprocess
import sys
from typing import Iterable

from thonny import get_runner, ui_utils
from thonny.assistance import SubprocessProgramAnalyzer, add_program_analyzer
from thonny.running import get_frontend_python
import logging


class MyPyAnalyzer(SubprocessProgramAnalyzer):
    
    def start_analysis(self, main_file_path, imported_file_paths: Iterable[str]) -> None:
        
        args = [get_frontend_python(), "-m", 
                "mypy", 
                "--ignore-missing-imports",
                "--check-untyped-defs",
                "--warn-redundant-casts",
                "--show-column-numbers",
                main_file_path
                ] + list(imported_file_paths)
        
        # TODO: ignore "... need type annotation" messages
        
        from mypy.version import __version__
        try:
            ver = tuple(map(int, __version__.split(".")))
        except Exception:
            ver = (0, 470) # minimum required version
        
        if ver >= (0, 520):
            args.insert(3, "--no-implicit-optional")
             
        if ver >= (0, 590):
            args.insert(3, "--python-executable")
            args.insert(4, get_runner().get_executable()) 
                
        self._proc = ui_utils.popen_with_ui_thread_callback(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            on_completion=self._parse_and_output_warnings
        )
        
        
    def _parse_and_output_warnings(self, pylint_proc):
        out = pylint_proc.stdout.read()
        err = pylint_proc.stderr.read()
        if err:
            print(err, file=sys.stderr)
            #logging.getLogger("thonny").warning("MyPy: " + err)
        #print(out)
        warnings = []
        for line in out.splitlines():
            m = re.match(r"(.*?):(\d+):(\d+):(.*?):(.*)", line.strip())
            if m is not None:
                message = m.group(5).strip()
                if message == "invalid syntax":
                    continue # user will see this as Python error
                
                atts = {
                    "filename" : m.group(1),
                    "lineno" : int(m.group(2)),
                    "col_offset" : int(m.group(3))-1,
                    "kind" : m.group(4).strip(), # always "error" ?
                    "msg" : message,
                    "group" : "warnings",
                }
                # TODO: add better categorization and explanation
                atts["symbol"] = "mypy-" + atts["kind"]
                warnings.append(atts)
            else:
                # Without line number?
                m = re.match(r"(.*?): (.*?):(.*)", line.strip())
                if m is not None:
                    message = m.group(3).strip()
                    if message == "invalid syntax":
                        continue # user will see this as Python error
                    
                    atts = {
                        "filename" : m.group(1),
                        "kind" : m.group(2).strip(), 
                        "msg" : message,
                        "group" : "warnings",
                    }
                    atts["symbol"] = "mypy-" + atts["kind"]
                    warnings.append(atts)
                else:
                    logging.error("Can't parse MyPy line: " + line)

        
        self.completion_handler(self, warnings)

def load_plugin():
    add_program_analyzer(MyPyAnalyzer)



import os

from thonny.backend import get_vm
from thonny.common import BackendEvent

_last_pos = (None, None)
_notification_is_sent = False
_LAST_POS_SETTING = "dock_user_windows.last_position"

def on_configure(event):
    global _last_pos, _notification_is_sent
    pos = event.x, event.y
    if pos != _last_pos:
        get_vm().set_option(_LAST_POS_SETTING, pos)
    
    if not _notification_is_sent:
        get_vm().send_message(BackendEvent("UserWindowAppeared"))
        _notification_is_sent = True
        

def patch_tkinter(module):
    flag_name = "has_docking_patch"
    if getattr(module, flag_name, False):
        return
    
    original_constructor = module.Tk.__init__
    
    def patched_Tk_constructor(self, *args, **kw):
        original_constructor(self, *args, **kw)
        
        # move window to the same place it was previously
        last_pos = get_vm().get_option(_LAST_POS_SETTING)
        if isinstance(last_pos, tuple):
            self.geometry("+%d+%d" % last_pos)
            
        self.wm_attributes("-topmost", 1)
        #self.overrideredirect(1)
        
        # I'm using bind_class because turtle._Screen later overwrites the bind handler
        self.bind_class("Tk", "<Configure>", on_configure, True)
        
    
    module.Tk.__init__ = patched_Tk_constructor
    setattr(module, flag_name, True)

def patch_turtle(module):
    # Turtle needs more tweaking because it later overrides the position set in the Tk constructor
    turtle_config = getattr(module, "_CFG", None) 
    if isinstance(turtle_config, dict):
        last_pos = get_vm().get_option(_LAST_POS_SETTING)
        if isinstance(last_pos, tuple):
            turtle_config["leftright"], turtle_config["topbottom"] = last_pos 

def load_plugin():
    if os.environ.get("DOCK_USER_WINDOWS", "False").lower() == "true":
        vm = get_vm()
        vm.add_import_handler("tkinter", patch_tkinter)
        vm.add_import_handler("turtle", patch_turtle)



import ast
from textwrap import dedent

from thonny import ast_utils
from thonny.ast_utils import pretty


def test_single_assignment():
    check_marked_ast("x=1", """/=Module
    body=[...]
        0=Assign @ 1.0  -  1.3
            targets=[...]
                0=Name @ 1.0  -  1.1
                    id='x'
                    ctx=Store
            value=Num @ 1.2  -  1.3
                n=1""")


def test_simple_io_program():
    check_marked_ast("""age = int(input("Enter age: "))
if age > 18:
    print("Hi")
else:
    print("Hello!", end='')
    print("What's your name?")
""", 
"""/=Module
    body=[...]
        0=Assign @ 1.0  -  1.31
            targets=[...]
                0=Name @ 1.0  -  1.3
                    id='age'
                    ctx=Store
            value=Call @ 1.6  -  1.31
                func=Name @ 1.6  -  1.9
                    id='int'
                    ctx=Load
                args=[...]
                    0=Call @ 1.10  -  1.30
                        func=Name @ 1.10  -  1.15
                            id='input'
                            ctx=Load
                        args=[...]
                            0=Str @ 1.16  -  1.29
                                s='Enter age: '
                        keywords=[]
                keywords=[]
        1=If @ 2.0  -  6.30
            test=Compare @ 2.3  -  2.11
                left=Name @ 2.3  -  2.6
                    id='age'
                    ctx=Load
                ops=[...]
                    0=Gt
                comparators=[...]
                    0=Num @ 2.9  -  2.11
                        n=18
            body=[...]
                0=Expr @ 3.4  -  3.15
                    value=Call @ 3.4  -  3.15
                        func=Name @ 3.4  -  3.9
                            id='print'
                            ctx=Load
                        args=[...]
                            0=Str @ 3.10  -  3.14
                                s='Hi'
                        keywords=[]
            orelse=[...]
                0=Expr @ 5.4  -  5.27
                    value=Call @ 5.4  -  5.27
                        func=Name @ 5.4  -  5.9
                            id='print'
                            ctx=Load
                        args=[...]
                            0=Str @ 5.10  -  5.18
                                s='Hello!'
                        keywords=[...]
                            0=keyword
                                arg='end'
                                value=Str @ 5.24  -  5.26
                                    s=''
                1=Expr @ 6.4  -  6.30
                    value=Call @ 6.4  -  6.30
                        func=Name @ 6.4  -  6.9
                            id='print'
                            ctx=Load
                        args=[...]
                            0=Str @ 6.10  -  6.29
                                s="What's your name?"
                        keywords=[]""")

def test_two_trivial_defs():
    check_marked_ast("""def f():
    pass
def f():
    pass""", """/=Module
    body=[...]
        0=FunctionDef @ 1.0  -  2.8
            name='f'
            args=arguments
                args=[]
                vararg=None
                kwonlyargs=[]
                kw_defaults=[]
                kwarg=None
                defaults=[]
            body=[...]
                0=Pass @ 2.4  -  2.8
            decorator_list=[]
            returns=None
        1=FunctionDef @ 3.0  -  4.8
            name='f'
            args=arguments
                args=[]
                vararg=None
                kwonlyargs=[]
                kw_defaults=[]
                kwarg=None
                defaults=[]
            body=[...]
                0=Pass @ 4.4  -  4.8
            decorator_list=[]
            returns=None""")

def test_id_def():
    check_marked_ast("""def f(x):
    return x
""", """/=Module
    body=[...]
        0=FunctionDef @ 1.0  -  2.12
            name='f'
            args=arguments
                args=[...]
                    0=arg @ 1.6  -  1.7
                        arg='x'
                        annotation=None
                vararg=None
                kwonlyargs=[]
                kw_defaults=[]
                kwarg=None
                defaults=[]
            body=[...]
                0=Return @ 2.4  -  2.12
                    value=Name @ 2.11  -  2.12
                        id='x'
                        ctx=Load
            decorator_list=[]
            returns=None""")

def test_simple_while_program():
    check_marked_ast("""x = int(input("Enter number: "))

while x > 0:
    print(x)
    x -= 1
""", """/=Module
    body=[...]
        0=Assign @ 1.0  -  1.32
            targets=[...]
                0=Name @ 1.0  -  1.1
                    id='x'
                    ctx=Store
            value=Call @ 1.4  -  1.32
                func=Name @ 1.4  -  1.7
                    id='int'
                    ctx=Load
                args=[...]
                    0=Call @ 1.8  -  1.31
                        func=Name @ 1.8  -  1.13
                            id='input'
                            ctx=Load
                        args=[...]
                            0=Str @ 1.14  -  1.30
                                s='Enter number: '
                        keywords=[]
                keywords=[]
        1=While @ 3.0  -  5.10
            test=Compare @ 3.6  -  3.11
                left=Name @ 3.6  -  3.7
                    id='x'
                    ctx=Load
                ops=[...]
                    0=Gt
                comparators=[...]
                    0=Num @ 3.10  -  3.11
                        n=0
            body=[...]
                0=Expr @ 4.4  -  4.12
                    value=Call @ 4.4  -  4.12
                        func=Name @ 4.4  -  4.9
                            id='print'
                            ctx=Load
                        args=[...]
                            0=Name @ 4.10  -  4.11
                                id='x'
                                ctx=Load
                        keywords=[]
                1=AugAssign @ 5.4  -  5.10
                    target=Name @ 5.4  -  5.5
                        id='x'
                        ctx=Store
                    op=Sub
                    value=Num @ 5.9  -  5.10
                        n=1
            orelse=[]""")

def test_call_with_pos_and_kw_arg():
    check_marked_ast("""f(3, t=45)
""", """/=Module
    body=[...]
        0=Expr @ 1.0  -  1.10
            value=Call @ 1.0  -  1.10
                func=Name @ 1.0  -  1.1
                    id='f'
                    ctx=Load
                args=[...]
                    0=Num @ 1.2  -  1.3
                        n=3
                keywords=[...]
                    0=keyword
                        arg='t'
                        value=Num @ 1.7  -  1.9
                            n=45""")

def test_call_with_pos_star_kw():
    check_marked_ast("""f(3, *kala, t=45)
    """, 
    """/=Module
    body=[...]
        0=Expr @ 1.0  -  1.17
            value=Call @ 1.0  -  1.17
                func=Name @ 1.0  -  1.1
                    id='f'
                    ctx=Load
                args=[...]
                    0=Num @ 1.2  -  1.3
                        n=3
                    1=Starred @ 1.5  -  1.10
                        value=Name @ 1.6  -  1.10
                            id='kala'
                            ctx=Load
                        ctx=Load
                keywords=[...]
                    0=keyword
                        arg='t'
                        value=Num @ 1.14  -  1.16
                            n=45""")

def test_call_with_single_keyword():
    check_marked_ast("""fff(t=45)
""", """/=Module
    body=[...]
        0=Expr @ 1.0  -  1.9
            value=Call @ 1.0  -  1.9
                func=Name @ 1.0  -  1.3
                    id='fff'
                    ctx=Load
                args=[]
                keywords=[...]
                    0=keyword
                        arg='t'
                        value=Num @ 1.6  -  1.8
                            n=45""")

def test_call_without_arguments():
    check_marked_ast("""fff()
""", """/=Module
    body=[...]
        0=Expr @ 1.0  -  1.5
            value=Call @ 1.0  -  1.5
                func=Name @ 1.0  -  1.3
                    id='fff'
                    ctx=Load
                args=[]
                keywords=[]""")

def test_call_with_attribute_function_and_keyword_arg():
    check_marked_ast("""rida.strip().split(maxsplit=1)
""", """/=Module
    body=[...]
        0=Expr @ 1.0  -  1.30
            value=Call @ 1.0  -  1.30
                func=Attribute @ 1.0  -  1.18
                    value=Call @ 1.0  -  1.12
                        func=Attribute @ 1.0  -  1.10
                            value=Name @ 1.0  -  1.4
                                id='rida'
                                ctx=Load
                            attr='strip'
                            ctx=Load
                        args=[]
                        keywords=[]
                    attr='split'
                    ctx=Load
                args=[]
                keywords=[...]
                    0=keyword
                        arg='maxsplit'
                        value=Num @ 1.28  -  1.29
                            n=1""")

def test_del_from_list_with_integer():
    check_marked_ast("""del x[0]""", """/=Module
    body=[...]
        0=Delete @ 1.0  -  1.8
            targets=[...]
                0=Subscript @ 1.4  -  1.8
                    value=Name @ 1.4  -  1.5
                        id='x'
                        ctx=Load
                    slice=Index
                        value=Num @ 1.6  -  1.7
                            n=0
                    ctx=Del""")


def test_del_from_list_with_string():
    check_marked_ast("""del x["blah"]""", """/=Module
    body=[...]
        0=Delete @ 1.0  -  1.13
            targets=[...]
                0=Subscript @ 1.4  -  1.13
                    value=Name @ 1.4  -  1.5
                        id='x'
                        ctx=Load
                    slice=Index
                        value=Str @ 1.6  -  1.12
                            s='blah'
                    ctx=Del""")

def test_full_slice1():
    check_marked_ast("""blah[:]
""", """/=Module
    body=[...]
        0=Expr @ 1.0  -  1.7
            value=Subscript @ 1.0  -  1.7
                value=Name @ 1.0  -  1.4
                    id='blah'
                    ctx=Load
                slice=Slice
                    lower=None
                    upper=None
                    step=None
                ctx=Load""")

def test_full_slice2():
    check_marked_ast("""blah[::]
""", """/=Module
    body=[...]
        0=Expr @ 1.0  -  1.8
            value=Subscript @ 1.0  -  1.8
                value=Name @ 1.0  -  1.4
                    id='blah'
                    ctx=Load
                slice=Slice
                    lower=None
                    upper=None
                    step=None
                ctx=Load""")

def test_non_ascii_letters_with_calls_etc():
    check_marked_ast("""täpitähed = "täpitähed"
print(täpitähed["tšahh"])
pöhh(pöhh=3)
""", """/=Module
    body=[...]
        0=Assign @ 1.0  -  1.23
            targets=[...]
                0=Name @ 1.0  -  1.9
                    id='täpitähed'
                    ctx=Store
            value=Str @ 1.12  -  1.23
                s='täpitähed'
        1=Expr @ 2.0  -  2.25
            value=Call @ 2.0  -  2.25
                func=Name @ 2.0  -  2.5
                    id='print'
                    ctx=Load
                args=[...]
                    0=Subscript @ 2.6  -  2.24
                        value=Name @ 2.6  -  2.15
                            id='täpitähed'
                            ctx=Load
                        slice=Index
                            value=Str @ 2.16  -  2.23
                                s='tšahh'
                        ctx=Load
                keywords=[]
        2=Expr @ 3.0  -  3.12
            value=Call @ 3.0  -  3.12
                func=Name @ 3.0  -  3.4
                    id='pöhh'
                    ctx=Load
                args=[]
                keywords=[...]
                    0=keyword
                        arg='pöhh'
                        value=Num @ 3.10  -  3.11
                            n=3""")

def test_nested_binops():
    """http://bugs.python.org/issue18374"""
    check_marked_ast("1+2-3", """/=Module
    body=[...]
        0=Expr @ 1.0  -  1.5
            value=BinOp @ 1.0  -  1.5
                left=BinOp @ 1.0  -  1.3
                    left=Num @ 1.0  -  1.1
                        n=1
                    op=Add
                    right=Num @ 1.2  -  1.3
                        n=2
                op=Sub
                right=Num @ 1.4  -  1.5
                    n=3""")

def test_multiline_string():
    """http://bugs.python.org/issue18370"""
    check_marked_ast("""pass
blah = \"\"\"first
second
third\"\"\"
pass""",
    """/=Module
    body=[...]
        0=Pass @ 1.0  -  1.4
        1=Assign @ 2.0  -  4.8
            targets=[...]
                0=Name @ 2.0  -  2.4
                    id='blah'
                    ctx=Store
            value=Str @ 2.7  -  4.8
                s='first\\nsecond\\nthird'
        2=Pass @ 5.0  -  5.4""")

def check_marked_ast(source, expected_pretty_ast
                     #,expected_for_py_34=None
                     ):
    
    #if (sys.version_info[:2] == (3,4) 
    #    and expected_for_py_34 is not None):
    #    expected_pretty_ast = expected_for_py_34
    
    source = dedent(source)
    root = ast.parse(source)
    ast_utils.mark_text_ranges(root, source.encode("utf-8"))
    actual_pretty_ast = pretty(root)
    #print("ACTUAL", actual_pretty_ast)
    #print("EXPECTED", expected_pretty_ast)
    assert actual_pretty_ast.strip() == expected_pretty_ast.strip() 



import re
import tkinter as tk
from tkinter import ttk

from thonny import get_workbench
from thonny.ui_utils import SafeScrollbar


class OutlineView(ttk.Frame):
    def __init__(self, master):
        ttk.Frame.__init__(self, master)
        self._init_widgets()
        
        self._tab_changed_binding = get_workbench().get_editor_notebook().bind("<<NotebookTabChanged>>", self._update_frame_contents ,True)
        get_workbench().bind("Save", self._update_frame_contents, True)
        get_workbench().bind("SaveAs", self._update_frame_contents, True)
        get_workbench().bind_class("Text", "<<NewLine>>", self._update_frame_contents, True)
        
        self._update_frame_contents()
    
    def destroy(self):
        try:
            # Not sure if editor notebook is still living
            get_workbench().get_editor_notebook().unbind("<<NotebookTabChanged>>", self._tab_changed_binding)
        except Exception:
            pass
        self.vert_scrollbar["command"] = None
        ttk.Frame.destroy(self)
    
    def _init_widgets(self):
        #init and place scrollbar
        self.vert_scrollbar = SafeScrollbar(self, orient=tk.VERTICAL)
        self.vert_scrollbar.grid(row=0, column=1, sticky=tk.NSEW)

        #init and place tree
        self.tree = ttk.Treeview(self, yscrollcommand=self.vert_scrollbar.set)
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)
        self.vert_scrollbar['command'] = self.tree.yview

        #set single-cell frame
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        #init tree events
        self.tree.bind("<Double-Button-1>", self._on_double_click, "+")

        #configure the only tree column
        self.tree.column('#0', anchor=tk.W, stretch=True)
        # self.tree.heading('#0', text='Item (type @ line)', anchor=tk.W)
        self.tree['show'] = ('tree',)
        
        self._class_img = get_workbench().get_image("outline-class")
        self._method_img = get_workbench().get_image("outline-method")

    def _update_frame_contents(self, event=None):
        self._clear_tree()
        
        editor = get_workbench().get_editor_notebook().get_current_editor()
        if editor is None:
            return
        
        root = self._parse_source(editor.get_code_view().get_content())
        for child in root[2]:
            self._add_item_to_tree('', child)
    
    def _parse_source(self, source):
        #all nodes in format (parent, node_indent, node_children, name, type, linenumber)
        root_node = (None, 0, [], None, None, None) #name, type and linenumber not needed for root
        active_node = root_node

        lineno = 0
        for line in source.split('\n'):
            lineno += 1
            m = re.match(r'[ ]*[\w]{1}', line)
            if m:
                indent = len(m.group(0))
                while indent <= active_node[1]:
                    active_node = active_node[0]

                t = re.match(r'[ ]*(?P<type>(def|class){1})[ ]+(?P<name>[\w]+)', line)
                if t:
                    current = (active_node, indent, [], t.group('name'), t.group('type'), lineno)
                    active_node[2].append(current)
                    active_node = current
        
        return root_node


    #adds a single item to the tree, recursively calls itself to add any child nodes
    def _add_item_to_tree(self, parent, item):
        #create the text to be played for this item
        item_type = item[4]
        item_text = " " + item[3]
        
        if item_type == "class":
            image = self._class_img
        elif item_type == "def":
            image = self._method_img
        else:
            image = None
        
        #insert the item, set lineno as a 'hidden' value
        current = self.tree.insert(parent, 'end', text=item_text, values = item[5], image=image)

        for child in item[2]:
            self._add_item_to_tree(current, child)
        
    #clears the tree by deleting all items      
    def _clear_tree(self):
        for child_id in self.tree.get_children():
            self.tree.delete(child_id)

    #called when a double-click is performed on any items
    def _on_double_click(self, event):
        editor = get_workbench().get_editor_notebook().get_current_editor()
        if editor:
            code_view = editor.get_code_view() 
            lineno = self.tree.item(self.tree.focus())['values'][0]
            index = code_view.text.index(str(lineno) + '.0')
            code_view.text.see(index) #make sure that the double-clicked item is visible
            code_view.select_lines(lineno, lineno)
            
            get_workbench().event_generate("OutlineDoubleClick",
                item_text=self.tree.item(self.tree.focus(), option='text'))

def load_plugin() -> None: 
    get_workbench().add_view(OutlineView, "Outline", "ne")



# -*- coding: utf-8 -*-

import os
import tkinter as tk
from tkinter.messagebox import showerror
#from thonny.ui_utils import askstring TODO: doesn't work
from tkinter.simpledialog import askstring

from thonny import get_workbench, misc_utils
from thonny.base_file_browser import BaseFileBrowser


class MainFileBrowser(BaseFileBrowser):
    def __init__(self, master, show_hidden_files=False):
        BaseFileBrowser.__init__(self, master, show_hidden_files, "file.last_browser_folder")
        
        self.menu = tk.Menu(tk._default_root, tearoff=False)
        self.menu.add_command(label="Create new file", command=self.create_new_file)
        
        self.tree.bind('<3>', self.on_secondary_click, True)
        if misc_utils.running_on_mac_os():
            self.tree.bind('<2>', self.on_secondary_click, True)
            self.tree.bind('<Control-1>', self.on_secondary_click, True)
    
                
    def create_new_file(self):
        selected_path = self.get_selected_path()
        
        if not selected_path:
            return
        
        if os.path.isdir(selected_path):
            parent_path = selected_path
        else:
            parent_path = os.path.dirname(selected_path)
        
        
        initial_name = self.get_proposed_new_file_name(parent_path, ".py")
        name = askstring("File name",
                        "Provide filename",
                        initialvalue=initial_name,
                        #selection_range=(0, len(initial_name)-3)
                        )
        
        if not name:
            return
                        
        path = os.path.join(parent_path, name)
        
        if os.path.exists(path):
            showerror("Error", "The file '"+path+"' already exists")
        else:
            open(path, 'w').close()
        
        self.open_path_in_browser(path, True)
        get_workbench().get_editor_notebook().show_file(path)

    def get_proposed_new_file_name(self, folder, extension):
        base = "new_file"
        
        if os.path.exists(os.path.join(folder, base + extension)):
            i = 2
            
            while True:
                name = base + "_" + str(i) + extension
                path = os.path.join(folder, name)
                if os.path.exists(path):
                    i += 1
                else:
                    return name
        else:
            return base + extension
         
    def on_secondary_click(self, event):
        node_id = self.tree.identify_row(event.y)
        
        if node_id:
            self.tree.selection_set(node_id)
            self.tree.focus(node_id)
            self.menu.tk_popup(event.x_root, event.y_root)
    
    def on_double_click(self, event):
        path = self.get_selected_path()
        if os.path.isfile(path):
            get_workbench().get_editor_notebook().show_file(path)
            self.save_current_folder()
        elif os.path.isdir(path):
            self.refresh_tree(self.get_selected_node(), True)
    
    
    
def load_plugin() -> None: 
    get_workbench().set_default("file.last_browser_folder", None)
    get_workbench().add_view(MainFileBrowser, "Files", "nw")



DEEPSLEEP_RESET = 4
HARD_RESET = 2

class I2C:
    ''
    def init():
        pass

    def readfrom():
        pass

    def readfrom_into():
        pass

    def readfrom_mem():
        pass

    def readfrom_mem_into():
        pass

    def readinto():
        pass

    def scan():
        pass

    def start():
        pass

    def stop():
        pass

    def write():
        pass

    def writeto():
        pass

    def writeto_mem():
        pass

PWRON_RESET = 1

class Pin:
    ''
    AF1_TIM1 = 1
    AF1_TIM2 = 1
    AF2_TIM3 = 2
    AF2_TIM4 = 2
    AF2_TIM5 = 2
    AF3_TIM10 = 3
    AF3_TIM11 = 3
    AF3_TIM9 = 3
    AF4_I2C1 = 4
    AF4_I2C3 = 4
    AF5_SPI1 = 5
    AF5_SPI2 = 5
    AF7_USART1 = 7
    AF7_USART2 = 7
    AF8_USART6 = 8
    AF9_I2C3 = 9
    AF_OD = 18
    AF_PP = 2
    ALT = 2
    ALT_OPEN_DRAIN = 18
    ANALOG = 3
    IN = 0
    IRQ_FALLING = 270598144
    IRQ_RISING = 269549568
    OPEN_DRAIN = 17
    OUT = 1
    OUT_OD = 17
    OUT_PP = 1
    PULL_DOWN = 2
    PULL_NONE = 0
    PULL_UP = 1
    def af():
        pass

    def af_list():
        pass

    board = None
    cpu = None
    def debug():
        pass

    def dict():
        pass

    def gpio():
        pass

    def high():
        pass

    def init():
        pass

    def irq():
        pass

    def low():
        pass

    def mapper():
        pass

    def mode():
        pass

    def name():
        pass

    def names():
        pass

    def off():
        pass

    def on():
        pass

    def pin():
        pass

    def port():
        pass

    def pull():
        pass

    def value():
        pass

SOFT_RESET = 0

class SPI:
    ''
    LSB = 128
    MSB = 0
    def deinit():
        pass

    def init():
        pass

    def read():
        pass

    def readinto():
        pass

    def write():
        pass

    def write_readinto():
        pass


class Signal:
    ''
    def off():
        pass

    def on():
        pass

    def value():
        pass


class UART:
    ''
    CTS = 512
    RTS = 256
    def any():
        pass

    def deinit():
        pass

    def init():
        pass

    def read():
        pass

    def readchar():
        pass

    def readinto():
        pass

    def readline():
        pass

    def sendbreak():
        pass

    def write():
        pass

    def writechar():
        pass


class WDT:
    ''
    def feed():
        pass

WDT_RESET = 3
def bootloader():
    pass

def deepsleep():
    pass

def disable_irq():
    pass

def enable_irq():
    pass

def freq():
    pass

def idle():
    pass

def info():
    pass

mem16 = None
mem32 = None
mem8 = None
def reset():
    pass

def reset_cause():
    pass

def sleep():
    pass

def soft_reset():
    pass

def time_pulse_us():
    pass

def unique_id():
    pass



def localtime():
    pass

def mktime():
    pass

def sleep():
    pass

def sleep_ms():
    pass

def sleep_us():
    pass

def ticks_add():
    pass

def ticks_cpu():
    pass

def ticks_diff():
    pass

def ticks_ms():
    pass

def ticks_us():
    pass

def time():
    pass



argv = None
byteorder = 'little'
def exit():
    pass

implementation = None
maxsize = 2147483647
modules = None
path = None
platform = 'pyboard'
def print_exception():
    pass

stderr = None
stdin = None
stdout = None
version = '3.4.0'
version_info = None



import os

from thonny import get_workbench

_OPTION_NAME = "run.dock_user_windows"

def toggle_variable():
    var = get_workbench().get_variable(_OPTION_NAME)
    var.set(not var.get())
    update_environment()

def update_environment():
    os.environ["DOCK_USER_WINDOWS"] = str(get_workbench().get_option(_OPTION_NAME))

def on_window_appear(event):
    get_workbench().become_active_window(force=False)

def load_plugin():
    get_workbench().set_default(_OPTION_NAME, False)
    get_workbench().add_command("toggle_dock_user_windows", "run", 
                                "Dock user windows",
                                toggle_variable,
                                flag_name=_OPTION_NAME,
                                group=35)
    update_environment()
    
    get_workbench().bind("UserWindowAppeared", on_window_appear, True)



# -*- coding: utf-8 -*-

from tkinter import ttk

from thonny import get_runner, get_workbench
from thonny.common import InlineCommand
from thonny.memory import VariablesFrame


class VariablesView(VariablesFrame):
    # TODO: Indicate invalid state when program or debug command is running more than a second
    def __init__(self, master):
        super().__init__(master)
        
        ttk.Style().configure("Centered.TButton", justify="center")
        self.back_button = ttk.Button(self.tree, style="Centered.TButton", 
                                      text="Back to\ncurrent frame",
                                      command=self._handle_back_button,
                                      width=15)
        
        get_workbench().bind("BackendRestart", self._handle_backend_restart, True)
        get_workbench().bind("ToplevelResponse", self._handle_toplevel_response, True)
        #get_workbench().bind("DebuggerResponse", self._debugger_response, True)
        get_workbench().bind("get_frame_info_response", self._handle_frame_info_event, True)
        get_workbench().bind("get_globals_response", self._handle_get_globals_response, True)
        
        # records last info from progress messages 
        self._last_active_info = None
    
    def _update_back_button(self, visible):
        if visible:
            assert self._last_active_info is not None
            self.back_button.configure(text="Back to\n"
                                       + self._last_active_info[-1])
            self.back_button.place(relx=1, x=-5, y=5, anchor="ne")
        else:
            self.back_button.place_forget()
    
    def _handle_back_button(self):
        assert self._last_active_info is not None
        if len(self._last_active_info) == 2:
            self.show_globals(*self._last_active_info)
        else:
            assert len(self._last_active_info) == 4
            self.show_frame_variables(*self._last_active_info)
    
    def _handle_backend_restart(self, event):
        self._clear_tree()

    def _handle_get_globals_response(self, event):
        self.show_globals(event["globals"], event["module_name"])
    
    def _handle_toplevel_response(self, event):
        if "globals" in event:
            self.show_globals(event["globals"], "__main__")
        else:
            # MicroPython
            get_runner().send_command(InlineCommand("get_globals", module_name="__main__"))
    
    
    def show_globals(self, globals_, module_name, is_active=True):
        # TODO: update only if something has changed
        self.update_variables(globals_)
        if module_name == "__main__":
            self._set_tab_caption("Variables")
        else:
            self._set_tab_caption("Variables (%s)" % module_name)
        
        if is_active:
            self._last_active_info = (globals_, module_name)
            
        self._update_back_button(not is_active)
    
    def show_frame_variables(self, locals_, globals_, freevars, frame_name, is_active=True):
        # TODO: update only if something has changed
        actual_locals = {}
        nonlocals = {}
        for name in locals_:
            if name in freevars:
                nonlocals[name] = locals_[name]
            else:
                actual_locals[name] = locals_[name]
                
        groups = [("LOCALS", actual_locals), 
                  ("GLOBALS", globals_)]
        if nonlocals:
            groups.insert(1, ("NONLOCALS", nonlocals))
        
        self.update_variables(groups)
        self._set_tab_caption("Variables (%s)" % frame_name)
        
        if is_active:
            self._last_active_info = (locals_, globals_, freevars, frame_name)
            
        self._update_back_button(not is_active)
        
    def _handle_frame_info_event(self, frame_info):
        if frame_info.get("error"):
            "probably non-existent frame"
            return
        else:
            is_active = (frame_info["location"] == "stack"
                         or # same __main__ globals can be in different frames
                             (frame_info["code_name"] == "<module>"
                             and frame_info["module_name"] == "__main__"
                             and self._last_active_info[-1] == "__main__"
                             and self._last_active_info[0] == frame_info["globals"]))
            
            if frame_info["code_name"] == "<module>":
                self.show_globals(frame_info["globals"], 
                                  frame_info["module_name"],
                                  is_active)
            else:
                self.show_frame_variables(frame_info["locals"],
                                          frame_info["globals"],
                                          frame_info["freevars"],
                                          frame_info["code_name"],
                                          is_active)
                  
    
    def _set_tab_caption(self, text):
        # pylint: disable=no-member
        if self.hidden:
            return
        
        self.home_widget.master.tab(self.home_widget, text=text)
    
    

def load_plugin() -> None:
    get_workbench().add_view(VariablesView, "Variables", "e", default_position_key="AAA")



# -*- coding: utf-8 -*-

import tkinter as tk
from tkinter import ttk

from thonny import get_workbench, misc_utils
from thonny.ui_utils import select_sequence

#TODO - consider moving the cmd_find method to main class in order to pass the editornotebook reference
#TODO - logging
#TODO - instead of text.see method create another one which attempts to center the line where the text was found
#TODO - test on mac and linux

# Handles the find dialog display and the logic of searching.
#Communicates with the codeview that is passed to the constructor as a parameter.

_active_find_dialog = None

class FindDialog(tk.Toplevel): 
    
    last_searched_word = None
    
    def __init__(self, master):
        padx=15
        pady=15
        
        tk.Toplevel.__init__(self, master, takefocus=1, background="pink")
        main_frame = ttk.Frame(self)
        main_frame.grid(row=1, column=1, sticky="nsew")
        self.columnconfigure(1, weight=1)
        self.rowconfigure(1, weight=1)
        
        self.codeview = master
        
        #references to the current set of passive found tags e.g. all words that match the searched term but are not the active string
        self.passive_found_tags = set()
        self.active_found_tag = None    #reference to the currently active (centered) found string

        #a tuple containing the start and indexes of the last processed string
        #if the last action was find, then the end index is start index + 1
        #if the last action was replace, then the indexes correspond to the start
        #and end of the inserted word
        self.last_processed_indexes = None
        self.last_search_case = None    #case sensitivity value used during the last search
        
        #set up window display
        self.geometry("+%d+%d" % (master.winfo_rootx() + master.winfo_width() // 2,
                                  master.winfo_rooty() + master.winfo_height() // 2 - 150))

        self.title("Find & Replace")
        if misc_utils.running_on_mac_os():
            self.configure(background="systemSheetBackground")
        self.resizable(height=tk.FALSE, width=tk.FALSE)
        self.transient(master) 
        self.protocol("WM_DELETE_WINDOW", self._ok)
      
        #Find text label
        self.find_label = ttk.Label(main_frame, text="Find:")   
        self.find_label.grid(column=0, row=0, sticky="w", padx=(padx, 0), pady=(pady, 0))

        #Find text field
        self.find_entry_var = tk.StringVar()
        self.find_entry = ttk.Entry(main_frame, textvariable=self.find_entry_var)
        self.find_entry.grid(column=1, row=0, columnspan=2, padx=(0, 10), pady=(pady, 0))
        if FindDialog.last_searched_word is not None:
            self.find_entry.insert(0, FindDialog.last_searched_word)

        #Replace text label
        self.replace_label = ttk.Label(main_frame, text="Replace with:") 
        self.replace_label.grid(column=0, row=1, sticky="w", padx=(padx, 0))

        #Replace text field
        self.replace_entry = ttk.Entry(main_frame)
        self.replace_entry.grid(column=1, row=1, columnspan=2, padx=(0, 10))

        #Info text label (invisible by default, used to tell user that searched string was not found etc)
        self.infotext_label_var = tk.StringVar()
        self.infotext_label_var.set("")
        self.infotext_label = ttk.Label(main_frame, textvariable=self.infotext_label_var, foreground="red") #TODO - style to conf
        self.infotext_label.grid(column=0, row=2, columnspan=3, pady=3, padx=(padx, 0))

        #Case checkbox
        self.case_var = tk.IntVar()
        self.case_checkbutton = ttk.Checkbutton(main_frame,text="Case sensitive",variable=self.case_var)
        self.case_checkbutton.grid(column=0, row=3, padx=(padx, 0), pady=(0, pady))

        #Direction radiobuttons
        self.direction_var = tk.IntVar()
        self.up_radiobutton = ttk.Radiobutton(main_frame, text="Up", variable=self.direction_var, value=1)
        self.up_radiobutton.grid(column=1, row=3, pady=(0, pady))
        self.down_radiobutton = ttk.Radiobutton(main_frame, text="Down", variable=self.direction_var, value=2)
        self.down_radiobutton.grid(column=2, row=3, pady=(0, pady))
        self.down_radiobutton.invoke()

        #Find button - goes to the next occurrence
        self.find_button = ttk.Button(main_frame, text="Find", command=self._perform_find, default="active")
        self.find_button.grid(column=3, row=0, sticky=tk.W + tk.E, pady=(pady, 0), padx=(0, padx))
        self.find_button.config(state='disabled') 

        #Replace button - replaces the current occurrence, if it exists
        self.replace_button = ttk.Button(main_frame, text="Replace", command=self._perform_replace)
        self.replace_button.grid(column=3, row=1, sticky=tk.W + tk.E, padx=(0, padx))
        self.replace_button.config(state='disabled')

        #Replace + find button - replaces the current occurence and goes to next
        self.replace_and_find_button = ttk.Button(main_frame, text="Replace+Find", command=self._perform_replace_and_find) #TODO - text to resources
        self.replace_and_find_button.grid(column=3, row=2, sticky=tk.W + tk.E, padx=(0, padx))
        self.replace_and_find_button.config(state='disabled')  
 
        #Replace all button - replaces all occurrences
        self.replace_all_button = ttk.Button(main_frame, text="Replace all", command=self._perform_replace_all) #TODO - text to resources
        self.replace_all_button.grid(column=3, row=3, sticky=tk.W + tk.E, padx=(0, padx), pady=(0, pady))        
        if FindDialog.last_searched_word == None:
            self.replace_all_button.config(state='disabled') 

        #create bindings
        self.bind('<Escape>', self._ok)
        self.find_entry_var.trace('w', self._update_button_statuses)
        self.find_entry.bind("<Return>", self._perform_find, True)
        self.bind("<F3>", self._perform_find, True)
        self.find_entry.bind("<KP_Enter>", self._perform_find, True)

        self._update_button_statuses()
        
        global _active_find_dialog
        _active_find_dialog = self
        self.focus_set()

    def focus_set(self):
        self.find_entry.focus_set()
        self.find_entry.selection_range(0, tk.END)

    #callback for text modifications on the find entry object, used to dynamically enable and disable buttons
    def _update_button_statuses(self, *args):
        find_text = self.find_entry_var.get()
        if len(find_text) == 0:
            self.find_button.config(state='disabled')
            self.replace_and_find_button.config(state='disabled')
            self.replace_all_button.config(state='disabled')
        else:
            self.find_button.config(state='normal')
            self.replace_all_button.config(state='normal')
            if self.active_found_tag is not None: 
                self.replace_and_find_button.config(state='normal')

    #returns whether the next search is case sensitive based on the current value of the case sensitivity checkbox
    def _is_search_case_sensitive(self):
        return self.case_var.get() != 0

    #returns whether the current search is a repeat of the last searched based on all significant values
    def _repeats_last_search(self, tofind):
        return tofind == FindDialog.last_searched_word and self.last_processed_indexes is not None and self.last_search_case == self._is_search_case_sensitive()


    #performs the replace operation - replaces the currently active found word with what is entered in the replace field
    def _perform_replace(self):

        #nothing is currently in found status
        if self.active_found_tag == None:
            return

        #get the found word bounds
        del_start = self.active_found_tag[0]
        del_end = self.active_found_tag[1]

        #erase all tags - these would not be correct anyway after new word is inserted
        self._remove_all_tags()
        toreplace = self.replace_entry.get() #get the text to replace

        #delete the found word
        self.codeview.text.delete(del_start, del_end)
        #insert the new word
        self.codeview.text.insert(del_start, toreplace)
        #mark the inserted word boundaries 
        self.last_processed_indexes = (del_start, self.codeview.text.index("%s+%dc" % (del_start, len(toreplace))))

        get_workbench().event_generate("Replace",
            widget=self.codeview.text,
            old_text=self.codeview.text.get(del_start, del_end),
            new_text=toreplace)

    #performs the replace operation followed by a new find
    def _perform_replace_and_find(self):
        if self.active_found_tag == None:
            return
        self._perform_replace()
        self._perform_find()

    #replaces all occurences of the search string with the replace string
    def _perform_replace_all(self):

        tofind = self.find_entry.get()
        if len(tofind) == 0:
            self.infotext_label_var.set("Enter string to be replaced.")
            return
        
        toreplace = self.replace_entry.get()
   
        self._remove_all_tags()

        currentpos = 1.0
        end = self.codeview.text.index("end")

        while True:
            currentpos = self.codeview.text.search(tofind, currentpos, end, nocase = not self._is_search_case_sensitive()) 
            if currentpos == "":
                break

            endpos = self.codeview.text.index("%s+%dc" % (currentpos, len(tofind)))

            self.codeview.text.delete(currentpos, endpos)

            if toreplace != "":
                self.codeview.text.insert(currentpos, toreplace)
                
            currentpos = self.codeview.text.index("%s+%dc" % (currentpos, len(toreplace)))

        get_workbench().event_generate("ReplaceAll",
            widget=self.codeview.text,
            old_text=tofind,
            new_text=toreplace)
        
        
    def _perform_find(self, event=None):
        self.infotext_label_var.set("")    #reset the info label text
        tofind = self.find_entry.get() #get the text to find 
        if len(tofind) == 0:    #in the case of empty string, cancel
            return              #TODO - set warning text to info label?

        search_backwards = self.direction_var.get() == 1 #True - search backwards ('up'), False - forwards ('down')
        
        if self._repeats_last_search(tofind): #continuing previous search, find the next occurrence
            if search_backwards:
                search_start_index = self.last_processed_indexes[0]
            else:
                search_start_index = self.last_processed_indexes[1]
            
            if self.active_found_tag is not None:
                self.codeview.text.tag_remove("current_found", self.active_found_tag[0], self.active_found_tag[1])  #remove the active tag from the previously found string
                self.passive_found_tags.add((self.active_found_tag[0], self.active_found_tag[1]))                   #..and set it to passive instead
                self.codeview.text.tag_add("found", self.active_found_tag[0], self.active_found_tag[1])
        
        else: #start a new search, start from the current insert line position
            if self.active_found_tag is not None:
                self.codeview.text.tag_remove("current_found", self.active_found_tag[0], self.active_found_tag[1]) #remove the previous active tag if it was present
            for tag in self.passive_found_tags:
                self.codeview.text.tag_remove("found", tag[0], tag[1])                                            #and remove all the previous passive tags that were present
            search_start_index = self.codeview.text.index("insert")    #start searching from the current insert position
            self._find_and_tag_all(tofind)                             #set the passive tag to ALL found occurences
            FindDialog.last_searched_word = tofind                     #set the data about last search
            self.last_search_case = self._is_search_case_sensitive()       

        

        wordstart = self.codeview.text.search(tofind, search_start_index, backwards = search_backwards, forwards = not search_backwards, nocase = not self._is_search_case_sensitive()) #performs the search and sets the start index of the found string
        if len(wordstart) == 0:
            self.infotext_label_var.set("The specified text was not found!") #TODO - better text, also move it to the texts resources list
            self.replace_and_find_button.config(state='disabled')
            self.replace_button.config(state='disabled')            
            return
        
        self.last_processed_indexes = (wordstart, self.codeview.text.index("%s+1c" % wordstart)) #sets the data about last search      
        self.codeview.text.see(wordstart) #moves the view to the found index
        wordend = self.codeview.text.index("%s+%dc" % (wordstart, len(tofind))) #calculates the end index of the found string
        self.codeview.text.tag_add("current_found", wordstart, wordend) #tags the found word as active
        self.active_found_tag = (wordstart, wordend)
        self.replace_and_find_button.config(state='normal')
        self.replace_button.config(state='normal')

        get_workbench().event_generate("Find",
            widget=self.codeview.text,
            text=tofind,
            backwards=search_backwards,
            case_sensitive=self._is_search_case_sensitive())
        
     
    def _ok(self, event=None):
        """Called when the window is closed. responsible for handling all cleanup."""
        self._remove_all_tags()
        self.destroy()

        global _active_find_dialog
        _active_find_dialog = None
        

    #removes the active tag and all passive tags
    def _remove_all_tags(self):
        for tag in self.passive_found_tags:
            self.codeview.text.tag_remove("found", tag[0], tag[1]) #removes the passive tags

        if self.active_found_tag is not None:
            self.codeview.text.tag_remove("current_found", self.active_found_tag[0], self.active_found_tag[1]) #removes the currently active tag   

        self.active_found_tag = None
        self.replace_and_find_button.config(state='disabled')
        self.replace_button.config(state='disabled')        
        
    #finds and tags all occurences of the searched term
    def _find_and_tag_all(self, tofind, force=False): 
        #TODO - to be improved so only whole words are matched - surrounded by whitespace, parentheses, brackets, colons, semicolons, points, plus, minus

        if self._repeats_last_search(tofind) and not force:   #nothing to do, all passive tags already set
            return

        currentpos = 1.0
        end = self.codeview.text.index("end")

        #searches and tags until the end of codeview
        while True:
            currentpos = self.codeview.text.search(tofind, currentpos, end, nocase = not self._is_search_case_sensitive()) 
            if currentpos == "":
                break

            endpos = self.codeview.text.index("%s+%dc" % (currentpos, len(tofind)))
            self.passive_found_tags.add((currentpos, endpos))
            self.codeview.text.tag_add("found", currentpos, endpos)
            
            currentpos = self.codeview.text.index("%s+1c" % currentpos)

def load_plugin() -> None:
    def cmd_open_find_dialog():
        if _active_find_dialog is not None:
            _active_find_dialog.focus_set()
        else:
            editor = get_workbench().get_editor_notebook().get_current_editor()
            if editor:
                FindDialog(editor._code_view)

    
    def find_f3(event):
        if _active_find_dialog is None:
            cmd_open_find_dialog()
        else:
            _active_find_dialog._perform_find(event)
         
    get_workbench().add_command("OpenFindDialog", "edit", 'Find & Replace',
        cmd_open_find_dialog,
        default_sequence=select_sequence("<Control-f>", "<Command-f>"),
        extra_sequences=["<Control-Greek_phi>"],
        )
    
    get_workbench().bind("<F3>", find_f3, True)




class DHT11:
    ''
    def humidity():
        pass

    def measure():
        pass

    def temperature():
        pass


class DHT22:
    ''
    def humidity():
        pass

    def measure():
        pass

    def temperature():
        pass


class DHTBase:
    ''
    def measure():
        pass

def dht_readinto():
    pass




class BytesIO:
    ''
    def close():
        pass

    def flush():
        pass

    def getvalue():
        pass

    def read():
        pass

    def readinto():
        pass

    def readline():
        pass

    def seek():
        pass

    def write():
        pass


class FileIO:
    ''
    def close():
        pass

    def flush():
        pass

    def read():
        pass

    def readinto():
        pass

    def readline():
        pass

    def readlines():
        pass

    def seek():
        pass

    def tell():
        pass

    def write():
        pass


class StringIO:
    ''
    def close():
        pass

    def flush():
        pass

    def getvalue():
        pass

    def read():
        pass

    def readinto():
        pass

    def readline():
        pass

    def seek():
        pass

    def write():
        pass


class TextIOWrapper:
    ''
    def close():
        pass

    def flush():
        pass

    def read():
        pass

    def readinto():
        pass

    def readline():
        pass

    def readlines():
        pass

    def seek():
        pass

    def tell():
        pass

    def write():
        pass

def open():
    pass



import os.path
import subprocess
import tkinter as tk
from shutil import which
from tkinter import messagebox, ttk

from thonny import get_workbench, running, ui_utils
from thonny.common import actual_path
from thonny.misc_utils import running_on_mac_os, running_on_windows
from thonny.plugins.backend_config_page import BackendDetailsConfigPage
from thonny.running import WINDOWS_EXE
from thonny.ui_utils import (SubprocessDialog, askdirectory, askopenfilename,
                             create_string_var)


class CustomCPythonConfigurationPage(BackendDetailsConfigPage):
    def __init__(self, master):
        super().__init__(master)
        
        self._configuration_variable = create_string_var(
            get_workbench().get_option("CustomInterpreter.path"))
        
        entry_label = ttk.Label(self, text="Which interpreter to use for running programs?")
        entry_label.grid(row=0, column=0, columnspan=2, sticky=tk.W)
        
        self._entry = ttk.Combobox(self,
                              exportselection=False,
                              textvariable=self._configuration_variable,
                              values=self._get_interpreters())
        
        self._entry.grid(row=1, column=0, columnspan=2, sticky=tk.NSEW)
        self._entry.state(['!disabled', 'readonly'])
        
        another_label = ttk.Label(self, text="Your interpreter isn't in the list?")
        another_label.grid(row=2, column=0, columnspan=2, sticky=tk.W, pady=(10,0))
        
        ttk.Style().configure("Centered.TButton", justify="center")
        self._select_button = ttk.Button(self, style="Centered.TButton",
                                         text="Locate another "
                                         + ("python.exe ..." if running_on_windows() else "python executable ...")
                                         + "\nNB! Thonny only supports Python 3.5 and later",
                                         command=self._select_executable)
        
        self._select_button.grid(row=3, column=0, columnspan=2, sticky=tk.NSEW)
        
        self._venv_button = ttk.Button(self, style="Centered.TButton",
                                         text="Create new virtual environment ...\n"
                                         + "(Select existing or create a new empty directory)", 
                                         command=self._create_venv)
        
        self._venv_button.grid(row=4, column=0, columnspan=2, sticky=tk.NSEW, pady=(5,0))
        
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
    
    def _select_executable(self):
        # TODO: get dir of current interpreter
        options = {"master" : self}
        if running_on_windows():
            options["filetypes"] = [('Python interpreters', 'python.exe'), ('all files', '.*')]
            
        filename = askopenfilename(**options)
        
        if filename:
            self._configuration_variable.set(filename)
    
    def _create_venv(self):
        path = None
        while True:
            path = askdirectory(master=self,
                               initialdir=path, 
                               title="Select empty directory for new virtual environment")
            if not path:
                return
            
            if os.listdir(path):
                messagebox.showerror("Bad directory", "Selected directory is not empty.\nSelect another or cancel.")
            else:
                break
        assert os.path.isdir(path)
        path = actual_path(path)
        
        proc = subprocess.Popen([running.get_frontend_python(), "-m", "venv", path],
                                stdin=None, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                universal_newlines=True)
        dlg = SubprocessDialog(self, proc, "Creating virtual environment")
        ui_utils.show_dialog(dlg)
        
        if running_on_windows():
            exe_path = actual_path(os.path.join(path, "Scripts", "python.exe"))
        else:
            exe_path = os.path.join(path, "bin", "python3")
        
        if os.path.exists(exe_path):
            self._configuration_variable.set(exe_path)
        
    
    def _get_interpreters(self):
        result = set()
        
        if running_on_windows():
            # registry
            result.update(self._get_interpreters_from_windows_registry())
            
            for minor in [5,6,7,8]:
                for dir_ in [
                            "C:\\Python3%d" % minor,
                            "C:\\Python3%d-32" % minor,
                            "C:\\Python3%d-64" % minor,
                            "C:\\Program Files\\Python 3.%d" % minor,
                            "C:\\Program Files\\Python 3.%d-64" % minor,
                            "C:\\Program Files (x86)\\Python 3.%d" % minor,
                            "C:\\Program Files (x86)\\Python 3.%d-32" % minor,
                            ]:
                    path = os.path.join(dir_, WINDOWS_EXE)
                    if os.path.exists(path):
                        result.add(actual_path(path))  
                
            # other locations
            for dir_ in ["C:\\Anaconda3",
                         os.path.expanduser("~/Anaconda3")]:
                path = os.path.join(dir_, WINDOWS_EXE)
                if os.path.exists(path):
                    result.add(actual_path(path))  
        
        else:
            # Common unix locations
            dirs = ["/bin", "/usr/bin", "/usr/local/bin",
                    os.path.expanduser("~/.local/bin")]
            for dir_ in dirs:
                # if the dir_ is just a link to another dir_, skip it
                # (not to show items twice)
                # for example on Fedora /bin -> usr/bin
                apath = actual_path(dir_)
                if apath != dir_ and apath in dirs:
                    continue
                for name in ["python3", "python3.5", "python3.6", "python3.7", "python3.8"]:
                    path = os.path.join(dir_, name)
                    if os.path.exists(path):
                        result.add(path)  
        
        if running_on_mac_os():
            for version in ["3.5", "3.6", "3.7", "3.8"]:
                dir_ = os.path.join("/Library/Frameworks/Python.framework/Versions",
                                    version, "bin")
                path = os.path.join(dir_, "python3")
                
                if os.path.exists(path):
                    result.add(path)
        
        for command in ["python3", "python3.5", "python3.5", "python3.6", "python3.7", "python3.8"]:
            path = which(command)
            if path is not None and os.path.isabs(path):
                result.add(path)
        
        for path in get_workbench().get_option("CustomInterpreter.used_paths"):
            if os.path.exists(path):
                result.add(actual_path(path))
        
        return sorted(result)
    
    def _get_interpreters_from_windows_registry(self):
        # https://github.com/python/cpython/blob/master/Tools/msi/README.txt
        import winreg
        result = set()
        for key in [winreg.HKEY_LOCAL_MACHINE,
                    winreg.HKEY_CURRENT_USER]:
            for version in ["3.5", "3.5-32", "3.5-64",
                            "3.6", "3.6-32", "3.6-64",
                            "3.7", "3.7-32", "3.7-64",
                            "3.8", "3.8-32", "3.8-64",
                            ]:
                try:
                    for subkey in [
                        'SOFTWARE\\Python\\PythonCore\\' + version + '\\InstallPath',
                        'SOFTWARE\\Python\\PythonCore\\Wow6432Node\\' + version + '\\InstallPath'
                                 ]:
                        dir_ = winreg.QueryValue(key, subkey)
                        if dir_:
                            path = os.path.join(dir_, WINDOWS_EXE)
                            if os.path.exists(path):
                                result.add(path)
                except Exception:
                    pass
        
        return result
    
    def should_restart(self):
        return self._configuration_variable.modified  # pylint: disable=no-member
        
    def apply(self):
        if not self.should_restart():
            return
        
        path = self._configuration_variable.get()
        if os.path.isfile(path):
            get_workbench().set_option("CustomInterpreter.path", path)



def alloc_emergency_exception_buf():
    pass

def const():
    pass

def heap_lock():
    pass

def heap_unlock():
    pass

def kbd_intr():
    pass

def mem_info():
    pass

def opt_level():
    pass

def qstr_info():
    pass

def schedule():
    pass

def stack_use():
    pass



import os.path
import sys

from setuptools import find_packages, setup

if sys.version_info < (3,5):
    raise RuntimeError("Thonny requires Python 3.5 or later")

setupdir = os.path.dirname(__file__)

with open(os.path.join(setupdir, 'thonny', 'VERSION'), encoding="ASCII") as f:
    version = f.read().strip()

requirements = []
for line in open(os.path.join(setupdir, 'requirements.txt'), encoding="ASCII"):
    if line.strip() and not line.startswith('#'):
        requirements.append(line)

setup(
      name="thonny",
      version=version,
      description="Python IDE for beginners",
      long_description="Thonny is a simple Python IDE with features useful for learning programming. See http://thonny.org for more info.",
      url="https://thonny.org",
      author="Aivar Annamaa and others",
      author_email="thonny@googlegroups.com",
      license="MIT",
      classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: MacOS X",
        "Environment :: Win32 (MS Windows)",
        "Environment :: X11 Applications",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "License :: Freeware",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Education",
        "Topic :: Software Development",
        "Topic :: Software Development :: Debuggers",
        "Topic :: Text Editors",
      ],
      keywords="IDE education debugger",
      project_urls={
          'Source code': 'https://bitbucket.org/plas/thonny',
          'Bug tracker': 'https://bitbucket.org/plas/thonny/issues',
      },
      platforms=["Windows", "macOS", "Linux"],
      install_requires=requirements,
      python_requires=">=3.5",
      packages=find_packages(),
      package_data={'': ['VERSION',  'res/*'],
                    'thonny.plugins.help' : ['*.rst'],
                    'thonny.plugins.micropython' : ['api_stubs/*.*'],
                    },
      entry_points={
        'gui_scripts': [
            'thonny = thonny:launch',
        ]
      },      
)



import io
import tokenize

from thonny import get_workbench
from thonny.codeview import CodeViewText
from thonny.shell import ShellText

_OPENERS = {')': '(', ']': '[', '}': '{'}

class ParenMatcher:

    def __init__(self, text):
        self.text = text
        self._update_scheduled = False
    
    def schedule_update(self):
        def perform_update():
            try:
                self.update_highlighting()
            finally:
                self._update_scheduled = False
        
        if not self._update_scheduled:
            self._update_scheduled = True
            self.text.after_idle(perform_update)

    def update_highlighting(self):
        self.text.tag_remove("surrounding_parens", "0.1", "end")
        self.text.tag_remove("unclosed_expression", "0.1", "end")

        if get_workbench().get_option("view.paren_highlighting"):
            self._update_highlighting_for_active_range()
    
    def _update_highlighting_for_active_range(self):
        start_index = "1.0"
        end_index = self.text.index("end")
        remaining = self._highlight_surrounding(start_index, end_index)
        self._highlight_unclosed(remaining, start_index, end_index)
    
    def _highlight_surrounding(self, start_index, end_index):
        open_index, close_index, remaining = self.find_surrounding(start_index, end_index)
        if None not in [open_index, close_index]:
            self.text.tag_add("surrounding_parens", open_index)
            self.text.tag_add("surrounding_parens", close_index)
        
        return remaining

    # highlights an unclosed bracket
    def _highlight_unclosed(self, remaining, start_index, end_index):
        # anything remaining in the stack is an unmatched opener
        # since the list is ordered, we can highlight everything starting from the first element
        if len(remaining) > 0:
            opener = remaining[0]
            open_index = "%d.%d" % (opener.start[0], opener.start[1])
            self.text.tag_add("unclosed_expression", open_index, end_index) 
    
    def _get_paren_tokens(self, source):
        result = []
        try: 
            tokens = tokenize.tokenize(io.BytesIO(source.encode('utf-8')).readline)
            for token in tokens:
                if token.string != "" and token.string in "()[]{}":
                    result.append(token)
        except Exception:
            # happens eg when parens are unbalanced or there is indentation error or ...
            pass
        
        return result

    def find_surrounding(self, start_index, end_index):
                
        stack = []
        opener, closer = None, None
        open_index, close_index = None, None
        
        start_row, start_col = map(int, start_index.split(".")) 
        source = self.text.get(start_index, end_index)
        
        # prepend source with empty lines and spaces to make 
        # token rows and columns match with widget indices
        source = ("\n" * (start_row-1)) + (" "*start_col) + source 
        
        for t in self._get_paren_tokens(source):
            if t.string == "" or t.string not in "()[]{}":
                continue
            if t.string in "([{":
                stack.append(t)
            elif len(stack) > 0:
                if stack[-1].string != _OPENERS[t.string]:
                    continue
                if not closer:
                    opener = stack.pop()
                    open_index = "%d.%d" % (opener.start[0], opener.start[1])
                    token_index = "%d.%d" % (t.start[0], t.start[1])
                    if self._is_insert_between_indices(open_index, token_index):
                        closer = t
                        close_index = token_index
                else:
                    stack.pop()
        
        return open_index, close_index, stack
        

    def _is_insert_between_indices(self, index1, index2):
        return self.text.compare("insert", ">=", index1) and \
               self.text.compare("insert-1c", "<=", index2)

class ShellParenMatcher(ParenMatcher):
    def _update_highlighting_for_active_range(self):
    
        # TODO: check that cursor is in this range
        index_parts = self.text.tag_prevrange("command", "end")
        
        if index_parts:
            start_index, end_index = index_parts
            remaining = self._highlight_surrounding(start_index, end_index)
            self._highlight_unclosed(remaining, start_index, "end")
            
def update_highlighting(event=None):
    text = event.widget
    if not hasattr(text, "paren_matcher"):
        if isinstance(text, CodeViewText):
            text.paren_matcher = ParenMatcher(text)
        elif isinstance(text, ShellText):
            text.paren_matcher = ShellParenMatcher(text)
        else:
            return
    
    text.paren_matcher.schedule_update()

def load_plugin() -> None:
    wb = get_workbench()  
    
    wb.set_default("view.paren_highlighting", True)
    wb.bind_class("CodeViewText", "<<CursorMove>>", update_highlighting, True)
    wb.bind_class("CodeViewText", "<<TextChange>>", update_highlighting, True)
    wb.bind_class("ShellText", "<<CursorMove>>", update_highlighting, True)
    wb.bind_class("ShellText", "<<TextChange>>", update_highlighting, True)
    wb.bind("<<UpdateAppearance>>", update_highlighting, True)
